"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from abc import ABCMeta
from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.context import Context
from nagini_translation.lib.program_nodes import (
    PythonClass,
    PythonExceptionHandler,
    PythonGlobalVar,
    PythonIOOperation,
    PythonMethod,
    PythonModule,
    PythonTryBlock,
    PythonType,
    PythonVar,
)
from nagini_translation.lib.jvmaccess import JVM
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Field,
    VarDecl,
    Predicate,
    Position,
    Stmt,
    StmtsAndExpr,
)
from nagini_translation.lib.typeinfo import TypeInfo
from nagini_translation.lib.util import (
    get_surrounding_try_blocks,
    InvalidProgramException,
    UnsupportedException
)
from nagini_translation.lib.viper_ast import ViperAST
from typing import List, Tuple, Union


class TranslatorConfig:
    """
    Contains the configuration of the translator, i.e. all the parts
    (specialized translates) it consists of.
    """

    def __init__(self, translator: 'Translator'):
        self.expr_translator = None
        self.stmt_translator = None
        self.call_translator = None
        self.contract_translator = None
        self.perm_translator = None
        self.pure_translator = None
        self.type_translator = None
        self.pred_translator = None
        self.io_operation_translator = None
        self.obligation_translator = None
        self.prog_translator = None
        self.method_translator = None
        self.type_factory = None
        self.translator = translator


class AbstractTranslator(metaclass=ABCMeta):
    """
    Abstract class which all specialized translators extend. Provides a number
    of interface methods through which specialized translators can interact, and
    forwards calls to those methods to the respective translators.
    """

    def __init__(self, config: TranslatorConfig, jvm: JVM, source_file: str,
                 type_info: TypeInfo, viper_ast: ViperAST) -> None:
        self.config = config
        self.viper = viper_ast
        self.jvm = jvm

    @property
    def type_factory(self):
        return self.config.type_factory

    @property
    def translator(self):
        return self.config.translator

    def translate_expr(self, node: ast.AST, ctx: Context,
                       target_type: object = None,
                       impure: bool = False,
                       as_read: bool = False) -> StmtsAndExpr:
        return self.config.expr_translator.translate_expr(
            node, ctx, target_type, impure, as_read)

    def translate_stmt(self, node: ast.AST, ctx: Context) -> List[Stmt]:
        return self.config.stmt_translator.translate_stmt(node, ctx)

    def translate_contract(self, node: ast.AST, ctx: Context) -> Expr:
        return self.config.contract_translator.translate_contract(node, ctx)

    def translate_perm(self, node: ast.AST, ctx: Context) -> Expr:
        return self.config.perm_translator.translate_perm(node, ctx)

    def get_arp_for_context(self, node: ast.AST, ctx: Context) -> Expr:
        return self.config.perm_translator.get_arp_for_context(node, ctx)

    def translate_exprs(self, nodes: List[ast.AST],
                        function: PythonMethod, ctx: Context) -> Expr:
        return self.config.pure_translator.translate_exprs(nodes, function, ctx)

    def get_type(self, node: ast.AST, ctx: Context) -> PythonClass:
        return self.config.type_translator.get_type(node, ctx)

    def translate_type(self, cls: PythonClass,
                       ctx: Context) -> 'silver.ast.Type':
        return self.config.type_translator.translate_type(cls, ctx)

    def translate_Call(self, node: ast.Call, ctx: Context, impure=False,
                       statement=False) -> StmtsAndExpr:
        return self.config.call_translator.translate_Call(node, ctx, impure, statement)

    def translate_constructor_call(self, target_class: PythonClass,
                                   node: ast.Call, args: List, arg_stmts: List,
                                   ctx: Context) -> StmtsAndExpr:
        return self.config.call_translator.translate_constructor_call(
            target_class, node, args, arg_stmts, ctx)

    def translate_predicate(self, pred: PythonMethod,
                            ctx: Context) -> 'ast.silver.Predicate':
        return self.config.pred_translator.translate_predicate(pred, ctx)

    def translate_static_field_access(self, field: PythonGlobalVar,
                                      receiver: Union[Expr, PythonType],
                                      node, ctx: Context) -> Expr:
        return self.config.expr_translator.translate_static_field_access(field, receiver,
                                                                         node, ctx)

    def translate_global_var_reference(self, target: PythonGlobalVar, node: ast.AST,
                                       ctx: Context) -> Expr:
        return self.config.expr_translator.translate_global_var_reference(target, node,
                                                                          ctx)

    def translate_io_operation(
            self,
            operation: PythonIOOperation,
            ctx: Context,
            ) -> Tuple[
                'ast.silver.Predicate',
                List['ast.silver.Function'],
                List['ast.silver.Method'],
                ]:
        return self.config.io_operation_translator.translate_io_operation(
            operation,
            ctx)

    def translate_method(self, method: PythonMethod,
                         ctx: Context) -> 'silver.ast.Method':
        return self.config.method_translator.translate_method(method, ctx)

    def translate_main_method(self, modules: List[PythonModule],
                              ctx: Context) -> 'silver.ast.Method':
        return self.config.method_translator.translate_main_method(modules, ctx)

    def translate_function(self, func: PythonMethod,
                           ctx: Context) -> 'silver.ast.Function':
        return self.config.method_translator.translate_function(func, ctx)

    def translate_predicate_family(self, root: PythonMethod,
            preds: List[PythonMethod], ctx: Context) -> 'ast.silver.Predicate':
        return self.config.pred_translator.translate_predicate_family(root,
                                                                      preds,
                                                                      ctx)

    def translate_operator(self, left: Expr, right: Expr, left_type: PythonType,
                           right_type: PythonType, node: ast.AST,
                           ctx: Context) -> StmtsAndExpr:
        return self.config.expr_translator.translate_operator(
            left, right, left_type, right_type, node, ctx)

    def create_exception_catchers(self, var: PythonVar,
                                  try_blocks: List[PythonTryBlock],
                                  call: ast.Call, ctx: Context) -> List[Stmt]:
        return self.config.expr_translator.create_exception_catchers(var,
                                                                     try_blocks,
                                                                     call, ctx)

    def extract_contract(self, method: PythonMethod, errorvarname: str,
                         is_constructor: bool,
                         ctx: Context) -> Tuple[List[Expr], List[Expr]]:
        return self.config.method_translator.extract_contract(method,
                                                              errorvarname,
                                                              is_constructor,
                                                              ctx)

    def inline_method(self, method: PythonMethod, args: List[PythonVar],
                      result_var: PythonVar, error_var: PythonVar,
                      ctx: Context) -> List[Stmt]:
        return self.config.call_translator.inline_method(method, args,
                                                         result_var, error_var,
                                                         ctx)

    def translate_contractfunc_call(self, node: ast.Call, ctx: Context,
                                    impure=False, statement=False) -> StmtsAndExpr:
        return self.config.contract_translator.translate_contractfunc_call(node, ctx,
                                                                           impure,
                                                                           statement)

    def translate_normal_call(self, target: PythonMethod, arg_stmts: List[Stmt],
                              args: List[Expr], arg_types: List[PythonType],
                              node: ast.AST, ctx: Context, impure=False) -> StmtsAndExpr:
        return self.config.call_translator.translate_normal_call(target, arg_stmts, args,
                                                                 arg_types, node, ctx,
                                                                 impure)

    def translate_normal_call_node(self, node: ast.Call, ctx: Context,
                                   impure=False) -> StmtsAndExpr:
        return self.config.call_translator.translate_normal_call_node(node, ctx, impure)

    def translate_io_contractfunc_call(self, node: ast.Call,
                                       ctx: Context) -> StmtsAndExpr:
        translator = self.config.io_operation_translator
        return translator.translate_io_contractfunc_call(node, ctx)

    def translate_io_operation_call(self, node: ast.Call,
                                    ctx: Context) -> StmtsAndExpr:
        translator = self.config.io_operation_translator
        return translator.translate_io_operation_call(node, ctx)

    def is_io_existential_defining_equality(self, node: ast.Expr,
                                            ctx: Context) -> bool:
        translator = self.config.io_operation_translator
        return translator.is_io_existential_defining_equality(node, ctx)

    def define_io_existential(self, node: ast.Expr, ctx: Context) -> None:
        translator = self.config.io_operation_translator
        translator.define_io_existential(node, ctx)

    def translate_get_ghost_output(self, node: ast.Expr,
                                   ctx: Context) -> List[Stmt]:
        translator = self.config.io_operation_translator
        return translator.translate_get_ghost_output(node, ctx)

    def translate_obligation_contractfunc_call(self, node: ast.Call,
                                               ctx: Context) -> StmtsAndExpr:
        translator = self.config.obligation_translator
        return translator.translate_obligation_contractfunc_call(node, ctx)

    def translate_must_invoke_token(self, node: ast.Call,
                                    ctx: Context) -> StmtsAndExpr:
        translator = self.config.obligation_translator
        return translator.translate_must_invoke_token(node, ctx)

    def translate_must_invoke_ctoken(self, node: ast.Call,
                                     ctx: Context) -> StmtsAndExpr:
        translator = self.config.obligation_translator
        return translator.translate_must_invoke_ctoken(node, ctx)

    def get_obligation_preamble(
            self,
            ctx: Context) -> Tuple[List[Predicate], List[Field]]:
        translator = self.config.obligation_translator
        return translator.get_obligation_preamble(ctx)

    def is_wait_level_comparison(self, node: ast.Compare,
                                 ctx: Context) -> bool:
        translator = self.config.obligation_translator
        return translator.is_wait_level_comparison(node, ctx)

    def translate_wait_level_comparison(self, node: ast.Compare,
                                        ctx: Context) -> StmtsAndExpr:
        translator = self.config.obligation_translator
        return translator.translate_wait_level_comparison(node, ctx)

    def create_level_call(self, expr: sil.RefExpression) -> sil.PermExpression:
        translator = self.config.obligation_translator
        return translator.create_level_call(expr)

    def create_level_below(
            self, expr: sil.PermExpression,
            residue_level_var: sil.PermExpression,
            ctx: Context) -> sil.BoolExpression:
        translator = self.config.obligation_translator
        return translator.create_level_below(expr, residue_level_var, ctx)

    def initialize_current_wait_level(
            self, current_wait_level: sil.PermExpression,
            residue_level_var: sil.PermExpression,
            ctx: Context) -> sil.BoolExpression:
        translator = self.config.obligation_translator
        return translator.initialize_current_wait_level(
            current_wait_level, residue_level_var, ctx)

    def create_method_node(
            self, ctx: Context, name: str,
            args: List[VarDecl], returns: List[VarDecl],
            pres: List[Expr], posts: List[Expr],
            locals: List[VarDecl], body: List[Stmt],
            position: Position, info: Info,
            method: PythonMethod = None,
            overriding_check: bool = False) -> List[Stmt]:
        translator = self.config.obligation_translator
        return translator.create_method_node(
            ctx, name, args, returns, pres, posts, locals, body,
            position, info, method, overriding_check)

    def create_method_call_node(
            self, ctx: Context, methodname: str, args: List[Expr],
            targets: List[Expr], position: Position, info: Info,
            target_method: PythonMethod = None,
            target_node: ast.Call = None) -> List[Stmt]:
        translator = self.config.obligation_translator
        return translator.create_method_call_node(
            ctx, methodname, args, targets, position, info, target_method,
            target_node)

    def create_method_fork(self, ctx: Context, targets, thread: Expr,
                           position: Position, info: Info,
                           target_node: ast.Call = None) -> List[Stmt]:
        translator = self.config.obligation_translator
        return translator.create_method_fork(ctx, targets, thread, position, info,
                                             target_node)

    def enter_loop_translation(
            self, node: Union[ast.While, ast.For], ctx: Context,
            err_var: PythonVar = None) -> None:
        translator = self.config.obligation_translator
        return translator.enter_loop_translation(node, ctx, err_var)

    def leave_loop_translation(self, ctx: Context) -> None:
        translator = self.config.obligation_translator
        return translator.leave_loop_translation(ctx)

    def create_while_node(
            self, ctx: Context, cond: Expr,
            invariants: List[Expr],
            locals: List[VarDecl],
            body: Stmt, node: Union[ast.While, ast.For]) -> List[Stmt]:
        translator = self.config.obligation_translator
        return translator.create_while_node(
            ctx, cond, invariants, locals, body, node)

    def translate_handler(self, handler: PythonExceptionHandler,
                          ctx: Context) -> List[Stmt]:
        return self.config.method_translator.translate_handler(handler, ctx)

    def translate_finally(self, block: PythonTryBlock,
                          ctx: Context) -> List[Stmt]:
        return self.config.method_translator.translate_finally(block, ctx)

    def type_check(self, lhs: Expr, type: PythonType,
                   position: 'silver.ast.Position', ctx: Context,
                   inhale_exhale: bool=True) -> Expr:
        return self.config.type_translator.type_check(
            lhs, type, position, ctx, inhale_exhale=inhale_exhale)

    def bind_type_vars(self, method: PythonMethod, ctx: Context) -> None:
        return self.config.method_translator.bind_type_vars(method, ctx)

    def create_tuple(self, vals: List[Expr], val_types: List[PythonType],
                     node: ast.AST, ctx: Context) -> Expr:
        return self.config.expr_translator.create_tuple(vals, val_types, node,
                                                        ctx)

    def translate_string(self, s: str, node: ast.AST, ctx: Context) -> Expr:
        return self.config.expr_translator.translate_string(s, node, ctx)

    def translate_args(self, target: PythonMethod, arg_nodes: List,
                       keywords: List, node: ast.AST, ctx: Context,
                       implicit_receiver=None) -> Tuple[List[Stmt], List[Expr],
                                                        List[PythonType]]:
        return self.config.call_translator.translate_args(target, arg_nodes,
                                                          keywords, node, ctx,
                                                          implicit_receiver)




"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

from nagini_translation.sif.translators.call import (
    SIFCallTranslator
)
from nagini_translation.sif.translators.contract import (
    SIFContractTranslator
)
from nagini_translation.sif.translators.expression import (
    SIFExpressionTranslator
)
from nagini_translation.sif.translators.method import (
    SIFMethodTranslator
)
from nagini_translation.sif.translators.program import (
    SIFProgramTranslator
)
from nagini_translation.sif.translators.statement import (
    SIFStatementTranslator
)

from nagini_translation.lib.jvmaccess import JVM
from nagini_translation.lib.typeinfo import TypeInfo
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.translator import Translator
from nagini_translation.translators.abstract import TranslatorConfig
from nagini_translation.translators.io_operation import IOOperationTranslator
from nagini_translation.translators.obligation import ObligationTranslator
from nagini_translation.translators.permission import PermTranslator
from nagini_translation.translators.predicate import PredicateTranslator
from nagini_translation.translators.pure import PureTranslator
from nagini_translation.translators.type import TypeTranslator
from nagini_translation.translators.type_domain_factory import (
    TypeDomainFactory
)


class SIFTranslator(Translator):
    """
    Translator producing extended Silver AST.
    """
    def __init__(self, jvm: JVM, source_file: str, type_info: TypeInfo,
                 viper_ast: ViperAST):
        config = TranslatorConfig(self)
        config.pure_translator = PureTranslator(config, jvm, source_file,
                                                type_info, viper_ast)
        config.call_translator = SIFCallTranslator(
            config, jvm, source_file, type_info, viper_ast)
        config.contract_translator = SIFContractTranslator(config, jvm,
                                                           source_file,
                                                           type_info,
                                                           viper_ast)
        config.expr_translator = SIFExpressionTranslator(config, jvm,
                                                         source_file,
                                                         type_info, viper_ast)
        config.pred_translator = PredicateTranslator(config, jvm, source_file,
                                                     type_info, viper_ast)
        config.io_operation_translator = IOOperationTranslator(
            config, jvm, source_file, type_info, viper_ast)
        config.obligation_translator = ObligationTranslator(
            config, jvm, source_file, type_info, viper_ast)
        config.stmt_translator = SIFStatementTranslator(config, jvm,
                                                        source_file,
                                                        type_info, viper_ast)
        config.perm_translator = PermTranslator(config, jvm, source_file,
                                                type_info, viper_ast)
        config.type_translator = TypeTranslator(config, jvm, source_file,
                                                type_info, viper_ast)
        config.prog_translator = SIFProgramTranslator(config, jvm, source_file,
                                                      type_info, viper_ast)
        config.method_translator = SIFMethodTranslator(config, jvm, source_file,
                                                       type_info, viper_ast)
        config.type_factory = TypeDomainFactory(viper_ast, self)

        self.obligation_translator = config.obligation_translator
        self.prog_translator = config.prog_translator
        self.expr_translator = config.expr_translator



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""IO operation termination checks."""


import ast

from typing import List

from nagini_translation.lib.context import Context
from nagini_translation.lib.errors import Rules, rules
from nagini_translation.lib.guard_collectors import (
    GuardCollectingVisitor,
)
from nagini_translation.lib.program_nodes import PythonIOOperation
from nagini_translation.lib.typedefs import (
    Expr,
    Stmt,
)
from nagini_translation.lib.util import (
    join_expressions,
)
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.translators.io_operation.common import (
    IOOperationCommonTranslator,
)


class TerminationCheckGenerator(GuardCollectingVisitor):
    """Class responsible for generating IO operation termination checks."""

    def __init__(self, io_translator: IOOperationCommonTranslator,
                 ctx: Context,
                 termination_condition: Expr,
                 termination_measure: Expr,
                 termination_measure_node: ast.AST) -> None:
        super().__init__()
        self._io_translator = io_translator
        self._ctx = ctx
        self._termination_condition = termination_condition
        self._termination_measure = termination_measure
        self._termination_measure_node = termination_measure_node
        self._current_operation = None          # type: PythonIOOperation
        self._current_operation_node = None     # type: ast.Call
        self._current_identifier = None         # type: str
        self._current_guard_condition = None    # type: Expr
        self._checks = []                       # type: List[Stmt]

    def create_checks(self, node: Expr = None) -> List[Stmt]:
        """Create assertions that check termination."""
        self._check_termination_measure_positive()
        if node is not None:
            self.traverse(node)
        return self._checks

    def _is_io_operation(self, node: ast.Call) -> bool:
        return isinstance(self._io_translator.get_target(node, self._ctx),
                          PythonIOOperation)

    def visit_Call(self, node: ast.Call) -> None:
        if self._is_io_operation(node):
            self._create_termination_checks(node)
        else:
            super().visit_Call(node)

    def _create_termination_checks(self, node: ast.Call) -> None:
        """Create termination checks for sub-operation."""
        assert self._current_operation is None
        assert self._current_identifier is None
        assert self._current_guard_condition is None
        assert self._current_operation_node is None

        self._current_operation_node = node

        operation = self._io_translator.get_target(node, self._ctx)
        self._current_operation = operation

        identifier = "{} ({}:{})".format(
            operation.name, node.lineno, node.col_offset)
        self._current_identifier = identifier

        self._current_guard_condition = self._create_guard_condition()

        self._check_gap()
        with self._ctx.aliases_context():
            self._io_translator.set_up_io_operation_input_aliases(
                self._current_operation, node, self._ctx)
            self._check_termination_condition()
            self._check_termination_measure()

        self._current_operation = None
        self._current_identifier = None
        self._current_guard_condition = None
        self._current_operation_node = None

    def _create_guard_condition(self) -> Expr:
        """Generate a Silver expression that guards current AST node."""
        guard_sil_parts = []
        for part in self.current_guard:
            sil_part = self._translate_expr(part, target_type=self._viper.Bool)
            guard_sil_parts.append(sil_part)
        and_operator = (
            lambda left, right:
            self._viper.And(left, right,
                            self._position(), self._no_info()))
        condition = join_expressions(
            and_operator, [self._termination_condition] + guard_sil_parts)
        return condition

    def _add_check(self, condition: Expr, comment_template: str,
                   position: 'viper_ast.IdentifierPosition') -> None:
        check = self._viper.Implies(
            self._current_guard_condition, condition,
            position, self._no_info())
        comment = comment_template.format(self._current_identifier)
        assertion = self._viper.Assert(
            check, position, self._to_info(comment))
        self._checks.append(assertion)

    def _check_termination_measure_positive(self) -> None:
        """Check that termination measure is positive."""
        self._current_guard_condition = self._termination_condition
        self._current_operation_node = self._termination_measure_node

        position = self._position(
            rules.TERMINATION_CHECK_MEASURE_NON_POSITIVE)
        positive = self._viper.GtCmp(
            self._termination_measure,
            self._viper.IntLit(0, position, self._no_info()),
            position, self._no_info())
        self._add_check(positive, "Termination measure must be positive.",
                        position)

        self._current_guard_condition = None
        self._current_operation_node = None

    def _check_gap(self) -> None:
        """Check that ``gap_io`` is disabled under termination condition."""
        if self._current_operation.name == 'gap_io':
            position = self._position(
                rules.TERMINATION_CHECK_GAP_ENABLED)
            false = self._viper.FalseLit(position, self._no_info())
            self._add_check(false, "Gap at {}.", position)

    def _check_termination_condition(self) -> None:
        """Check that child termination condition is implied."""
        termination_condition = self._translate_expr(
            self._current_operation.get_terminates(),
            target_type=self._viper.Bool)
        position = self._position(
            rules.TERMINATION_CHECK_CHILD_TERMINATION_NOT_IMPLIED)
        self._add_check(termination_condition,
                        "Termination condition of {}.",
                        position)

    def _check_termination_measure(self) -> None:
        """Check that child measure is strictly smaller."""
        termination_measure = self._translate_expr(
            self._current_operation.get_termination_measure(),
            target_type=self._viper.Int)
        position = self._position(
            rules.TERMINATION_CHECK_MEASURE_NON_DECREASING)
        larger = self._viper.GtCmp(
            self._termination_measure,
            termination_measure,
            position, self._no_info())
        self._add_check(larger, "Termination measure of {}.", position)

    def _translate_expr(self, node: ast.AST, target_type=None) -> Expr:
        statement, expression = self._io_translator.translate_expr(
            node, self._ctx, target_type=target_type)
        assert not statement
        return expression

    def _no_info(self) -> 'viper_ast.NoInfo':
        return self._io_translator.no_info(self._ctx)

    def _to_info(self, comment) -> 'viper_ast.SimpleInfo':
        return self._io_translator.to_info([comment], self._ctx)

    def _position(
            self,
            conversion_rules: Rules=None) -> 'viper_ast.IdentifierPosition':
        return self._io_translator.to_position(
            self._current_operation_node, self._ctx, rules=conversion_rules)

    @property
    def _viper(self) -> ViperAST:
        return self._io_translator.viper



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Common code for both IO operation use and definition translation."""


import ast

from typing import List, Optional

from nagini_translation.lib.constants import (
    BOXED_PRIMITIVES,
    PRIMITIVE_PREFIX,
)
from nagini_translation.lib.context import Context
from nagini_translation.lib.program_nodes import (
    PythonIOOperation,
    PythonIOExistentialVar,
    PythonVar,
)
from nagini_translation.lib.typedefs import Expr
from nagini_translation.translators.common import CommonTranslator
from nagini_translation.translators.io_operation.utils import (
    construct_getter_name,
)


class IOOperationCommonTranslator(CommonTranslator):
    """Shared code between IO operation use and definition translation."""

    def translate_args(
            self, args: List[ast.expr],
            params,
            ctx: Context) -> List[Expr]:
        """Translate IO operation arguments to silver."""
        arg_exprs = []
        for arg, param in zip(args, params):
            typ = self.translate_type(param.type, ctx)
            arg_stmt, arg_expr = self.translate_expr(arg, ctx, target_type=typ)
            assert not arg_stmt
            arg_exprs.append(arg_expr)
        return arg_exprs

    def create_result_getter(
            self, node: ast.Call, result: PythonVar, ctx: Context,
            sil_args: List[ast.Expr] = None,
            operation: PythonIOOperation = None) -> Expr:
        """Construct a getter for an IO operation result."""
        position = self.no_position(ctx)
        info = self.no_info(ctx)

        if not operation:
            operation = self.get_target(node, ctx)
            assert isinstance(operation, PythonIOOperation)

        if sil_args is None:
            parameters = operation.get_parameters()
            py_args = node.args[:len(parameters)]
            sil_args = self.translate_args(py_args, parameters, ctx)

        getter_name = construct_getter_name(operation, result)
        typ = self.translate_type(result.type, ctx)
        formal_args = [
            arg.decl
            for arg in operation.get_parameters()
        ]

        getter = self.viper.FuncApp(
            getter_name, sil_args, position, info, typ, formal_args)
        return getter

    def set_up_io_operation_input_aliases(
            self, operation: PythonIOOperation, node: ast.Call,
            ctx: Context) -> List[str]:
        """Set up aliases from operation's parameters to its arguments."""
        aliases = []

        parameters = operation.get_parameters()
        py_args = node.args[:len(parameters)]
        sil_args = self.translate_args(py_args, parameters, ctx)
        for parameter, py_arg, sil_arg in zip(parameters, py_args, sil_args):
            var_type = self.get_type(py_arg, ctx).try_unbox()
            var = PythonIOExistentialVar(parameter.name, py_arg, var_type)
            var.set_ref(sil_arg, None)
            ctx.set_alias(parameter.name, var)
            aliases.append(parameter.name)

        return aliases



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Integer typed Silver expressions."""


import ast

from typing import List, Optional, Union

from nagini_translation.lib.silver_nodes.expression import Expression
from nagini_translation.lib.silver_nodes.int_cmp_expr import (
    GtCmp,
    GeCmp,
    LtCmp,
    LeCmp,
)
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Position,
)
from nagini_translation.lib.util import (
    join_expressions,
)


def _auto_box(value: Union['IntExpression', int]) -> 'IntExpression':
    """If ``value`` is Python ``int``, wrap it."""
    if isinstance(value, int):
        return RawIntExpression(value)
    else:
        return value


class IntExpression(Expression):   # pylint: disable=abstract-method
    """A base class for all integer expressions."""

    def get_value(self) -> Optional[int]:
        """Try to extract Python integer representing this expression.

        Otherwise, return ``None``.
        """
        return None

    def __gt__(self, other: Union['IntExpression', int]) -> 'BoolExpression':
        return GtCmp(self, _auto_box(other))

    def __ge__(self, other: Union['IntExpression', int]) -> 'BoolExpression':
        return GeCmp(self, _auto_box(other))

    def __lt__(self, other: Union['IntExpression', int]) -> 'BoolExpression':
        return LtCmp(self, _auto_box(other))

    def __le__(self, other: Union['IntExpression', int]) -> 'BoolExpression':
        return LeCmp(self, _auto_box(other))

    def __add__(self, other: Union['IntExpression', int]) -> 'IntExpression':
        return Add(self, _auto_box(other))

    def __sub__(self, other: Union['IntExpression', int]) -> 'IntExpression':
        return Sub(self, _auto_box(other))


class PythonIntExpression(IntExpression):
    """An integer expression represented by Python int expression."""

    def __init__(self, node: ast.expr) -> None:
        self._node = node

    def get_value(self) -> int:
        if isinstance(self._node, ast.Num):
            return self._node.n
        else:
            return None

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        int_class = ctx.module.global_module.classes['int']
        assert translator.get_type(self._node, ctx) is int_class
        stmt, expr = translator.translate_expr(
            self._node, ctx, target_type=translator.viper.Int)
        assert not stmt
        return expr


class RawIntExpression(IntExpression):
    """Just a raw integer."""

    def __init__(self, value: int) -> None:
        self._value = value

    def get_value(self) -> int:
        return self._value

    def __add__(self, other: Union['IntExpression', int]) -> 'IntExpression':
        if isinstance(other, int):
            return RawIntExpression(self._value + other)
        return Add(self, other)

    def __sub__(self, other: Union['IntExpression', int]) -> 'IntExpression':
        if isinstance(other, int):
            return RawIntExpression(self._value - other)
        return Sub(self, other)

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        return translator.viper.IntLit(self._value, position, info)


class Add(IntExpression):
    """Add two integers."""

    def __init__(self, left: IntExpression, right: IntExpression) -> None:
        self._left = left
        self._right = right

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        left = self._left.translate(translator, ctx, position, info)
        right = self._right.translate(translator, ctx, position, info)
        return translator.viper.Add(left, right, position, info)


class Sub(IntExpression):
    """Subtract two integers."""

    def __init__(self, left: IntExpression, right: IntExpression) -> None:
        self._left = left
        self._right = right

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        left = self._left.translate(translator, ctx, position, info)
        right = self._right.translate(translator, ctx, position, info)
        return translator.viper.Sub(left, right, position, info)


class Sum(IntExpression):
    """A sum of 0 or more elements."""

    def __init__(self, elements: List[IntExpression]) -> None:
        self._elements = elements

    def is_empty(self) -> None:
        """Check if have any elements."""
        return not self._elements

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        if not self._elements:
            return translator.viper.IntLit(0, position, info)
        else:
            elements = [
                element.translate(translator, ctx, position, info)
                for element in self._elements]
            plus_operator = (
                lambda left, right:
                translator.viper.Add(left, right, position, info))
            return join_expressions(plus_operator, elements)


class Inc(IntExpression):
    """Some expression + 1."""

    def __init__(self, value: IntExpression) -> None:
        self._value = value

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        value = self._value.translate(translator, ctx, position, info)
        return translator.viper.Add(
            value, translator.viper.IntLit(1, position, info), position,
            info)


class CondInc(IntExpression):
    """Some expression + 1 if condition holds."""

    def __init__(
            self, condition: 'BoolExpression', value: IntExpression) -> None:
        self._condition = condition
        self._value = value
        self._increased_value = value + 1

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        condition = self._condition.translate(translator, ctx, position, info)
        value = self._value.translate(translator, ctx, position, info)
        increased_value = self._increased_value.translate(
            translator, ctx, position, info)
        return translator.viper.CondExp(
            condition, increased_value, value, position, info)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Code for constructing obligation inhales and exhales in contracts."""


import abc

from typing import List, Optional, Tuple

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.context import Context
from nagini_translation.lib.errors import Rules, rules


class ObligationInhaleExhale:
    """Class for constructing obligation inhales and exhales in contracts.

    .. note::

        Credit here has a different meaning than in the original paper
        about obligations. It is used for encoding ``ctoken``.
    """

    def __init__(
            self, bounded: sil.Location,
            unbounded: Optional[sil.Location] = None,
            credit: Optional[sil.Location] = None,
            skip_exhale: bool = False,
            skip_inhale: bool = False,
            credit_only: bool = False) -> None:
        """Constructor.

        :param skip_exhale:
            Indicates that exhale part should be equivalent to ``True``.
        """
        self._bounded = bounded
        self._unbounded = unbounded
        self._credit = credit
        self._skip_exhale = skip_exhale
        self._skip_inhale = skip_inhale
        self._credit_only = credit_only

    @property
    def _unbounded_positive(self) -> sil.BoolExpression:
        assert self._unbounded is not None
        return sil.CurrentPerm(self._unbounded) > sil.NoPerm()

    @property
    def _unbounded_acc(self) -> sil.Acc:
        assert self._unbounded is not None
        return sil.Acc(self._unbounded)

    @property
    def _bounded_positive(self) -> sil.BoolExpression:
        return sil.CurrentPerm(self._bounded) > sil.NoPerm()

    @property
    def _bounded_acc(self) -> sil.Acc:
        return sil.Acc(self._bounded)

    @property
    def _credit_acc(self) -> sil.Acc:
        assert self._credit is not None
        return sil.Acc(self._credit)

    def _construct_inhale(self, fresh: bool) -> sil.Acc:
        """Construct obligation inhale."""
        if self._skip_inhale:
            return sil.TrueLit()
        if self._credit_only:
            return self._credit_acc
        if fresh:
            return self._unbounded_acc
        else:
            return self._bounded_acc

    def _construct_unbounded_exhale(self) -> sil.BoolExpression:
        """Construct unbounded obligation exhale."""
        if self._unbounded is None:
            return sil.FalseLit()
        if self._credit is None:
            return self._unbounded_acc
        else:
            return sil.BoolCondExp(
                self._unbounded_positive,
                self._unbounded_acc,
                self._credit_acc)

    def _construct_bounded_exhale(
            self,
            check: Optional[sil.BoolExpression]) -> sil.BoolExpression:
        """Construct bounded obligation exhale.

        :param check: check if measure is positive
        """
        checks = [self._bounded_positive]
        if check is not None:
            checks.append(check)
        return sil.BoolCondExp(
            sil.BigAnd(checks),
            self._bounded_acc,
            self._construct_unbounded_exhale())

    def construct_use_method_unbounded(self) -> sil.BoolExpression:
        """Construct inhale exhale pair for use in method contract.

        Used for fresh obligations.
        """
        if self._credit_only:
            exhale = self._credit_acc
        else:
            exhale = self._construct_unbounded_exhale()
        return sil.InhaleExhale(self._construct_inhale(True), exhale)

    def construct_use_method_bounded(
            self, measure_check: sil.BoolExpression,
            is_postconditon: bool) -> sil.BoolExpression:
        """Construct inhale exhale pair for use in method contract.

        Used for bounded obligations.
        """
        if self._skip_exhale:
            exhale = sil.TrueLit()
        elif self._credit_only:
            exhale = self._credit_acc
        else:
            exhale = self._construct_bounded_exhale(
                measure_check if not is_postconditon else None)
        return sil.InhaleExhale(self._construct_inhale(False), exhale)

    def construct_use_loop(
            self, measure_check: Optional[sil.BoolExpression],
            loop_check_before_var: sil.BoolVar) -> sil.BoolExpression:
        """Construct inhale exhale pair for use in loop invariant.

        ``measure_check is None`` indicates that obligation is fresh.
        """
        if self._skip_exhale:
            exhale = sil.TrueLit()
        elif self._credit_only:
            exhale = self._credit_acc
        elif measure_check is None:
            exhale = self._construct_unbounded_exhale()
        else:
            exhale = sil.BoolCondExp(
                loop_check_before_var,
                self._construct_bounded_exhale(None),
                self._construct_bounded_exhale(measure_check))
        inhale = self._construct_inhale(measure_check is None)
        return sil.InhaleExhale(inhale, exhale)

    def construct_obligation_bound(self) -> sil.Statement:
        """Construct statement for bounding obligation."""
        if self._credit_only:
            return sil.Exhale(sil.TrueLit())
        return sil.If(
            self._unbounded_positive,
            [sil.Exhale(self._unbounded_acc), sil.Inhale(self._bounded_acc)],
            [])


class InexhaleObligationInstanceMixin(abc.ABC):
    """Mixin that provides obligation use methods for obligation instances.

    This mix-in should be mixed into ``ObligationInstance`` that desire
    to have default ``get_use_method`` and ``get_use_loop`` behaviour.
    """

    @abc.abstractmethod
    def _get_inexhale(self, ctx: Context) -> ObligationInhaleExhale:
        """Create ``ObligationInhaleExhale`` instance."""

    def get_use_method(
            self, ctx: Context) -> List[Tuple[sil.Expression, Rules]]:
        """Default implementation for obligation use in method contract."""
        inexhale = self._get_inexhale(ctx)
        obligation_info = ctx.actual_function.obligation_info
        if self.is_fresh():
            return [(inexhale.construct_use_method_unbounded(), None)]
        else:
            terms = []

            # Positive measure.
            if not ctx.obligation_context.is_translating_posts:
                positive_measure = self.get_measure() > 0
                if not positive_measure.is_always_true():
                    terms.append((positive_measure,
                                  rules.OBLIGATION_MEASURE_NON_POSITIVE))

            # Actual inhale / exhale.
            measure_check = obligation_info.caller_measure_map.check(
                self.get_target(), self.get_measure())
            inhale_exhale = inexhale.construct_use_method_bounded(
                measure_check,
                ctx.obligation_context.is_translating_posts)
            terms.append((inhale_exhale, None))

            return terms

    def get_use_loop(
            self, ctx: Context) -> List[Tuple[sil.Expression, Rules]]:
        """Default implementation for obligation use in loop invariant."""
        obligation_info = ctx.obligation_context.current_loop_info

        terms = []

        # Positive measure.
        if not self.is_fresh():
            loop_condition = obligation_info.construct_loop_condition()
            positive_measure = sil.Implies(
                loop_condition, self.get_measure() > 0)
            if not positive_measure.is_always_true():
                terms.append((positive_measure,
                              rules.OBLIGATION_LOOP_MEASURE_NON_POSITIVE))

        # Actual inhale / exhale.
        inexhale = self._get_inexhale(ctx)
        if self.is_fresh():
            measure_check = None
        else:
            measure_check = obligation_info.loop_measure_map.check(
                self.get_target(), self.get_measure())
        inhale_exhale = inexhale.construct_use_loop(
            measure_check, sil.BoolVar(obligation_info.loop_check_before_var))
        terms.append((inhale_exhale, None))

        return terms

    def get_obligation_bound(self, ctx: Context) -> sil.Statement:
        """Default implementation for bounding an obligation."""
        assert self.is_fresh()
        inexhale = self._get_inexhale(ctx)
        return inexhale.construct_obligation_bound()



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

DEFAULT_CLIENT_SOCKET = "tcp://localhost:5555"
DEFAULT_SERVER_SOCKET = "tcp://*:5555"


LITERALS = ['True', 'False', 'None']

BUILTINS = ['cast',
            'int',
            'isinstance',
            'bool',
            'len',
            'str',
            'set',
            'super',
            'range',
            'type',
            'list',
            'enumerate']

THREADING = ['Thread']

BUILTIN_PREDICATES = ['list_pred', 'set_pred', 'dict_pred', 'MayStart', 'ThreadPost']

FUNCTION_DOMAIN_NAME = 'Function'

MAY_SET_PRED = '_MaySet'

IS_DEFINED_FUNC = '_isDefined'

ASSERTING_FUNC = '_asserting'

NAME_QUANTIFIER_VAR = '_name'

COMBINE_NAME_FUNC = '_combine'

GLOBAL_IS_DEFINED_FUNC = '_isDefinedG'

CHECK_DEFINED_FUNC = '_checkDefined'

GLOBAL_CHECK_DEFINED_FUNC = '_checkDefinedG'

ARBITRARY_BOOL_FUNC = '_int_to_bool'

JOINABLE_FUNC = '_joinable'

THREAD_POST_PRED = '_thread_post'

THREAD_START_PRED = '_thread_start'

THREAD_DOMAIN = 'Thread'

METHOD_ID_DOMAIN = 'ThreadingID'

GET_ARG_FUNC = 'getArg'

GET_OLD_FUNC = 'getOld'

GET_METHOD_FUNC = 'getMethod'

GLOBAL_VAR_FIELD = '_val'

NAME_DOMAIN = '_Name'

COMBINED_NAME_ACCESSOR = '_get_combined_name'

COMBINED_PREFIX_ACCESSOR = '_get_combined_prefix'

SINGLE_NAME = '_single'

INTERNAL_NAMES = [
    'FuncTriple',
    'ft_get1',
    'ft_get2',
    'ft_get2',
    'ft_create',
    'A1',
    'A2',
    'A3',
    'PyType',
    'extends_',
    'issubtype',
    'isnotsubtype',
    'typeof',
    'get_type_arg1',
    'get_type_arg2',
    'get_type_nargs0',
    'get_type_nargs1',
    'issubtype_transitivity',
    'issubtype_reflexivity',
    'extends_implies_subtype',
    'issubtype_exclusion',
    'issubtype_exclusion_2',
    'issubtype_exclusion_propagation',
    'Thread',
    JOINABLE_FUNC,
    THREAD_POST_PRED,
    THREAD_START_PRED,
    METHOD_ID_DOMAIN,
    GET_ARG_FUNC,
    GET_OLD_FUNC,
    GET_METHOD_FUNC,
    MAY_SET_PRED,
    IS_DEFINED_FUNC,
    CHECK_DEFINED_FUNC,
    FUNCTION_DOMAIN_NAME,
    ARBITRARY_BOOL_FUNC,
    GLOBAL_VAR_FIELD,
    NAME_QUANTIFIER_VAR,
    COMBINE_NAME_FUNC,
    NAME_DOMAIN,
    COMBINED_NAME_ACCESSOR,
    COMBINED_PREFIX_ACCESSOR,
    SINGLE_NAME,
    'm',     # the following are used in various
    'X',     # places in the resources/... files.
    'Y',
    'id',
    't',
    'g',
    'x',
    'Low'
]

VIPER_KEYWORDS = [
    'result',
    'Int',
    'Perm',
    'Bool',
    'Ref',
    'Rational',
    'true',
    'false',
    'null',
    'import',
    'method',
    'function',
    'predicate',
    'program',
    'domain',
    'axiom',
    'var',
    'returns',
    'field',
    'define',
    'wand',
    'requires',
    'ensures',
    'invariant',
    'fold',
    'unfold',
    'inhale',
    'exhale',
    'new',
    'assert',
    'assume',
    'package',
    'apply',
    'elseif',
    'goto',
    'label',
    'fresh',
    'constraining',
    'Seq',
    'Set',
    'Multiset',
    'union',
    'intersection',
    'setminus',
    'subset',
    'unfolding',
    'in',
    'folding',
    'applying',
    'packaging',
    'old',
    'lhs',
    'let',
    'forall',
    'exists',
    'forperm',
    'acc',
    'wildcard',
    'write',
    'none',
    'epsilon',
    'perm',
    'unique']

LEGAL_MAGIC_METHODS = {
    '__eq__',
    '__ne__',
    '__gt__',
    '__ge__',
    '__lt__',
    '__le__',
    '__add__',
    '__sub_',
    '__mul__',
    '__floordiv__',
    '__mod__',
    '__init__',
    '__enter__',
    '__exit__',
    '__str__',
    '__len__',
    '__bool__',
    '__getitem__',
    '__setitem__',
}

RESULT_NAME = '_res'

ERROR_NAME = '_err'

END_LABEL = '__end'

LIST_TYPE = 'list'

RANGE_TYPE = 'range'

PSEQ_TYPE = 'PSeq'

PSET_TYPE = 'PSet'

PMSET_TYPE = 'PMultiset'

TUPLE_TYPE = 'tuple'

UNION_TYPE = 'Union'

DICT_TYPE = 'dict'

SET_TYPE = 'set'

STRING_TYPE = 'str'

BYTES_TYPE = 'bytes'

INT_TYPE = 'int'

FLOAT_TYPE = 'float'

BOOL_TYPE = 'bool'

PRIMITIVE_PREFIX = '__prim__'

PRIMITIVE_INT_TYPE = PRIMITIVE_PREFIX + INT_TYPE

PRIMITIVE_BOOL_TYPE = PRIMITIVE_PREFIX + BOOL_TYPE

PRIMITIVE_SEQ_TYPE = PRIMITIVE_PREFIX + 'Seq'

PRIMITIVE_SET_TYPE = PRIMITIVE_PREFIX + 'Set'

PRIMITIVE_MSET_TYPE = PRIMITIVE_PREFIX + 'Multiset'


OBJECT_TYPE = 'object'

CALLABLE_TYPE = 'Callable'

PRIMITIVES = {PRIMITIVE_INT_TYPE, PRIMITIVE_BOOL_TYPE, PRIMITIVE_SEQ_TYPE,
              PRIMITIVE_SET_TYPE, CALLABLE_TYPE, PRIMITIVE_MSET_TYPE}

BOXED_PRIMITIVES = {INT_TYPE, BOOL_TYPE}

NAME_VAR = '__name__'

FILE_VAR = '__file__'

MODULE_VARS = (NAME_VAR, FILE_VAR)

MAIN_METHOD_NAME = '__main__'

MYPY_SUPERCLASSES = {
    'Sized',
}

EVAL_IO_SIGNATURE = ('eval_io', 'func', 'arg', 'result')

IGNORED_IMPORTS = {'_importlib_modulespec',
                   'abc',
                   'builtins',
                   'nagini_contracts',
                   'nagini_contracts.contracts',
                   'nagini_contracts.io_contracts',
                   'nagini_contracts.obligations',
                   'nagini_contracts.thread',
                   'sys',
                   'types',
                   'typing',
                   }

IGNORED_MODULE_NAMES = {
    '_importlib_modulespec': [],
    'abc': [],
    'builtins': [],
    'nagini_contracts': [],
    'nagini_contracts.contracts': [],
    'nagini_contracts.io_contracts': [],
    'nagini_contracts.obligations': ['BaseLock'],
    'sys': [],
    'types': [],
    'typing': [],
}

OPERATOR_FUNCTIONS = {
    ast.Add: '__add__',
    ast.Sub: '__sub__',
    ast.Mult: '__mul__',
    ast.FloorDiv: '__floordiv__',
    ast.Div: '__div__',
    ast.Mod: '__mod__',
}



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""Silver built-in types."""


import abc

from typing import List

from nagini_translation.lib.typedefs import (
    Expr,
    Position,
    Info,
)


class Type:
    """A class for types."""

    @abc.abstractmethod
    def translate(self, translator: 'AbstractTranslator') -> Expr:
        """Translate type to its Silver representation."""

    def adjust_type(self, translator: 'AbstractTranslator', e: Expr,
                    ctx: 'Context') -> Expr:
        """Convert the expression e to this Silver type if possible."""
        return e


class BoolType(Type):
    """A boolean type."""

    def translate(self, translator: 'AbstractTranslator') -> Expr:
        return translator.viper.Bool

    def adjust_type(self, translator: 'AbstractTranslator', e: Expr,
                    ctx: 'Context') -> Expr:
        return translator.to_bool(e, ctx)


BOOL = BoolType()


class IntType(Type):
    """An integer type."""

    def translate(self, translator: 'AbstractTranslator') -> Expr:
        return translator.viper.Int

    def adjust_type(self, translator: 'AbstractTranslator', e: Expr,
                    ctx: 'Context') -> Expr:
        return translator.to_int(e, ctx)

INT = IntType()


class RefType(Type):
    """A reference type."""

    def translate(self, translator: 'AbstractTranslator') -> Expr:
        return translator.viper.Ref

    def adjust_type(self, translator: 'AbstractTranslator', e: Expr,
                    ctx: 'Context') -> Expr:
        return translator.to_ref(e, ctx)


REF = RefType()


class PermType(Type):
    """The Viper permission type."""

    def translate(self, translator: 'AbstractTranslator') -> Expr:
        return translator.viper.Perm


PERM = PermType()


class SeqType(Type):
    """A sequence type."""

    def __init__(self, element_type: Type) -> None:
        self._element_type = element_type

    def translate(self, translator: 'AbstractTranslator') -> Expr:
        element_type = self._element_type.translate(translator)
        return translator.viper.SeqType(element_type)


class DomainType(Type):
    """A domain type."""

    def __init__(self, name: str) -> None:
        self.name = name

    def translate(self, translator: 'AbstractTranslator') -> Expr:
        return translator.viper.DomainType(self.name, {}, [])


class DomainFuncApp:
    """A domain function call."""

    def __init__(
            self, name: str, args: List['Expression'],
            domain: DomainType) -> None:
        self._name = name
        self._args = args
        self._domain = domain

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        """Translate to Silver DomainFuncApp."""
        args = [arg.translate(translator, ctx, position, info)
                for arg in self._args]
        domain = self._domain.translate(translator)
        return translator.viper.DomainFuncApp(
            self._name, args, domain, position, info,
            self._domain.name)


class Domain:
    """A helper class for using a domain defined in Silver."""

    def __init__(self, name: str) -> None:
        self._name = name
        self._functions = set()

    def get_type(self) -> DomainType:
        """Return the Domain Type."""
        return DomainType(self._name)

    def declare_function(self, name: str) -> None:
        """Declare new function in domain."""
        self._functions.add(name)

    def call_function(
            self, name: str, args: List['Expression']) -> DomainFuncApp:
        """Call a domain function."""
        assert name in self._functions
        return DomainFuncApp(name, args, self.get_type())


class PSeq:
    """A helper class for generating Silver sequences."""

    def __init__(self, typ: Type, elements: List['Expression']) -> None:
        self._type = typ
        self._elements = elements

    def translate(self, translator: 'AbstractTranslator', ctx: 'Context',
                  position: Position, info: Info) -> Expr:
        """Translate to Silver sequence."""
        if not self._elements:
            typ = self._type.translate(translator)
            return translator.viper.EmptySeq(typ, position, info)
        else:
            elements = [element.translate(translator, ctx, position, info)
                        for element in self._elements]
            return translator.viper.ExplicitSeq(elements, position, info)



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast

from nagini_translation.lib.typedefs import Function, Method, Stmt
from nagini_translation.lib.program_nodes import PythonMethod, PythonModule
from nagini_translation.translators.abstract import Context
from nagini_translation.translators.method import MethodTranslator

from typing import List

class SIFMethodTranslator(MethodTranslator):
    """
    Extended AST version of method translator.
    """

    def get_error_var(self, stmt: ast.AST, ctx: Context) -> 'silver.ast.LocalVarRef':
        return ctx.current_function.error_var.ref()

    def _method_body_postamble(self, method: PythonMethod, ctx: Context) -> List[Stmt]:
        # With extended AST we don't need goto end or catch blocks.
        return []

    def _create_method_epilog(self, method: PythonMethod, ctx: Context) -> List[Stmt]:
        # With the extended AST we don't need a label at the end of the method.
        # Check that no undeclared exceptions are raised. (but not for main method)
        if not method.declared_exceptions:
            no_info = self.no_info(ctx)
            error_string = '"method raises no exceptions"'
            error_pos = self.to_position(method.node, ctx, error_string)
            assert_no_err = self.viper.AssertNoException(error_pos, no_info)
            return [assert_no_err]
        return []

    def translate_method(self, method: PythonMethod, ctx: Context) -> Method:
        self.viper.ctx = ctx
        self.viper.type_factory = self.type_factory
        if method.all_low:
            self.viper.all_low_methods.add(method.sil_name)
        elif method.preserves_low:
            self.viper.preserves_low_methods.add(method.sil_name)
        return super().translate_method(method, ctx)

    def translate_main_method(self, modules: List[PythonModule],
                              ctx: Context) -> List[Method]:
        no_pos = self.no_position(ctx)
        no_info = self.no_info(ctx)

        main = self._get_main_module(modules)
        main_method, locals, stmts = self._create_main_method_setup(modules, ctx)
        method_name = main_method.sil_name

        # Create an error variable
        error_var = self.create_method_error_var(ctx)
        main_method.error_var = error_var
        locals.append(error_var.decl)
        stmts.append(self.viper.LocalVarAssign(
            error_var.ref(),
            self.viper.NullLit(no_pos, no_info), no_pos, no_info))

        # Translate statements in main module. When an import statement is encountered,
        # the translation will include executing the statements in the imported module.
        for stmt in main.node.body:
            stmts.extend(self.translate_stmt(stmt, ctx))

        stmts += self._method_body_postamble(main_method, ctx)
        stmts += self._create_method_epilog(main_method, ctx)

        main_locals = [local.decl for local in main_method.get_locals()
                       if not local.name.startswith('lambda')]
        res = self.create_method_node(ctx, method_name, [], [], [], [],
                                      main_locals + locals, stmts, no_pos,
                                      no_info, method=ctx.current_function)
        ctx.current_function = None
        return main_method, res



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""A helper class for translating IO operation results."""


import ast
import copy

from typing import cast, List

from nagini_translation.lib.constants import EVAL_IO_SIGNATURE
from nagini_translation.lib.context import Context
from nagini_translation.lib.io_context import IOOpenContext
from nagini_translation.lib.program_nodes import (
    PythonIOOperation,
    PythonIOExistentialVar,
    PythonGlobalVar,
    PythonVar,
    PythonVarBase,
)
from nagini_translation.lib.typedefs import (
    Expr,
)
from nagini_translation.lib.viper_ast import ViperAST
from nagini_translation.translators.io_operation.common import (
    IOOperationCommonTranslator,
)
from nagini_translation.translators.io_operation.utils import (
    raise_invalid_operation_use,
    raise_invalid_existential_var,
)


class IOOperationResult:
    """Translation state of a specific IO operation result."""

    def __init__(self, node: ast.Call, definition: PythonVar,
                 instance: ast.expr) -> None:
        self._node = node

        self.definition = definition
        """Result as defined in IO operation definition."""

        self.instance = instance
        """Result as given in IO operation call."""

        self.getter = None
        """Getter associated with this result."""

        self.var_name = None
        """Variable name mentioned in IO operation call."""

        self._var = None

    @property
    def node(self) -> ast.Call:
        """Get the AST node for position."""
        return self._node

    @property
    def var(self) -> PythonVarBase:
        """The resolved variable that was referred by ``var_name``."""
        assert self._var is not None
        return self._var

    @var.setter
    def var(self, var: PythonVarBase) -> None:
        assert var is not None
        if var.type != self.definition.type:
            if not (self.var_name == EVAL_IO_SIGNATURE[3] and
                    self._node.func.id == EVAL_IO_SIGNATURE[0]):
                raise_invalid_existential_var(
                    'defining_expression_type_mismatch', self._node)
        self._var = var


class ResultTranslator:
    """Translate variables bound by IO operation call.

    .. todo:: Vytautas

        Check if it makes sense to have arbitrary expressions in result
        positions. (Currently, only variables are allowed.)
    """

    def __init__(
            self, operation: PythonIOOperation, node: ast.Call,
            translator: IOOperationCommonTranslator, ctx: Context) -> None:
        self._operation = operation
        self._node = node
        self._ctx = ctx
        self._translator = translator

        # Translated stuff.
        self._equations = []

    @property
    def _viper(self) -> ViperAST:
        return self._translator.viper

    @property
    def _position(self) -> 'viper_ast.Position':
        return self._translator.to_position(self._node, self._ctx)

    @property
    def _info(self) -> 'viper_ast.Info':
        return self._translator.no_info(self._ctx)

    @property
    def _io_ctx(self) -> IOOpenContext:
        return self._ctx.io_open_context

    @property
    def _result_definitions(self) -> List[PythonVar]:
        """Return results as defined in IO operation definition."""
        result = self._operation.get_results()
        if self._operation.name == EVAL_IO_SIGNATURE[0]:
            # Special treatment: Exchange result arg with one with the correct type.
            result = result[:]
            result_var = copy.copy(result[0])
            target = self._translator.get_target(self._node.args[1], self._ctx)
            if target.type.name != 'Callable':
                result_var.type = target.type
            result[0] = result_var
        return result

    @property
    def _result_instances(self) -> List[ast.expr]:
        """Return results as given in IO operation call."""
        parameters_count = len(self._operation.get_parameters())
        return self._node.args[parameters_count:]

    def translate(self) -> List[Expr]:
        """Translate variables bound by IO operation call."""
        assert not self._equations, "translate is called twice."

        self._check()
        contexts = self._create_result_contexts()

        for result in contexts:
            if not isinstance(result.instance, ast.Name):
                raise_invalid_operation_use(
                    'not_variable_in_result_position', self._node)
            else:
                result.getter = self._translator.create_result_getter(
                    self._node, result.definition, self._ctx)
                self._set_variable(result)

                if self._io_ctx.contains_variable(result.var_name):
                    # Variable denotes a result of the operation being opened.
                    self._handle_opener_result_var(result)
                elif isinstance(result.var, PythonIOExistentialVar):
                    self._handle_existential_var(result)
                else:
                    # Normal variable, which is already defined.
                    self._handle_normal_var(result)

        return self._equations

    def _check(self) -> None:
        """Perform some well-formedness checks."""
        if len(self._result_definitions) != len(self._result_instances):
            raise_invalid_operation_use('result_mismatch', self._node)

    def _create_result_contexts(self) -> List[IOOperationResult]:
        contexts = [
            IOOperationResult(self._node, definition, instance)
            for definition, instance in zip(
                self._result_definitions, self._result_instances)
        ]
        return contexts

    def _add_equation(self, result: IOOperationResult) -> None:
        equation = self._viper.EqCmp(result.getter,
                                     result.var.ref(result.node, self._ctx),
                                     self._position, self._info)
        self._equations.append(equation)

    def _set_variable(self, result: IOOperationResult) -> None:
        instance = cast(ast.Name, result.instance)
        var_name = instance.id
        result.var_name = var_name
        if var_name in self._ctx.var_aliases:
            result.var = self._ctx.var_aliases[var_name]
        else:
            result.var = self._ctx.actual_function.get_variable(var_name)

    def _handle_opener_result_var(
            self, result: IOOperationResult) -> None:
        assert isinstance(result.var, PythonVar)
        if self._io_ctx.is_variable_defined(result.var_name):
            self._add_equation(result)
        else:
            self._io_ctx.define_variable(result.var_name, result.getter)

    def _handle_existential_var(
            self, result: IOOperationResult) -> None:
        assert isinstance(result.var, PythonIOExistentialVar)
        if result.var.is_defined():
            self._add_equation(result)
        else:
            getter = result.getter
            old_getter = self._viper.Old(getter, getter.pos(), getter.info())
            result.var.set_ref(getter, old_getter)

    def _handle_normal_var(self, result: IOOperationResult) -> None:
        assert isinstance(result.var, (PythonVar, PythonGlobalVar))
        self._add_equation(result)



import ast
import copy
import operator
from typing import List

from nagini_translation.lib import silver_nodes as sil
from nagini_translation.lib.constants import (
    GET_ARG_FUNC,
    GET_METHOD_FUNC,
    GET_OLD_FUNC,
    JOINABLE_FUNC,
    METHOD_ID_DOMAIN,
    OBJECT_TYPE,
    THREAD_DOMAIN,
    THREAD_POST_PRED,
)
from nagini_translation.lib.context import Context
from nagini_translation.lib.program_nodes import PythonMethod
from nagini_translation.lib.typedefs import (
    Expr,
    Info,
    Position,
    Stmt
)
from nagini_translation.lib.util import (
    InvalidProgramException,
    OldExpressionCollector,
    OldExpressionTransformer,
    pprint,
)
from nagini_translation.translators.obligation.manager import (
    ObligationManager,
)
from nagini_translation.translators.obligation.node_constructor import (
    StatementNodeConstructorBase,
)


class ObligationMethodForkConstructor(StatementNodeConstructorBase):
    """A class that creates a method call node with obligation stuff."""

    def __init__(
            self, targets, thread: Expr,
            position: Position, info: Info,
            translator: 'AbstractTranslator', ctx: Context,
            obligation_manager: ObligationManager,
            target_node: ast.AST) -> None:
        super().__init__(
            translator, ctx, obligation_manager, position, info,
            target_node)
        self._statements = []
        self._targets = targets
        self._method_statements = {}
        self._thread = thread
        self._target_node = target_node
        self.viper = self._translator.viper

    def construct_fork(self) -> None:
        """Construct statements to perform a call."""
        self._add_precondition_exhales()
        self._add_waitlevel()

    def _create_join_permission(self, method: PythonMethod) -> Expr:
        """
        Creates an assertion representing a permission to join the thread and the
        ThreadPost predicate if the target method terminates.
        """
        tcond = method.obligation_info.create_termination_check(True)
        tcond = tcond.translate(self._translator, self._ctx, self._position,
                                self._info)

        joinable_func = self.viper.FuncApp(JOINABLE_FUNC, [self._thread],
                                           self._position, self._info,
                                           self.viper.Bool)
        post_pred_acc = self.viper.PredicateAccess([self._thread], THREAD_POST_PRED,
                                                   self._position, self._info)
        full_perm = self.viper.FullPerm(self._position, self._info)
        post_pred = self.viper.PredicateAccessPredicate(post_pred_acc, full_perm,
                                                        self._position, self._info)
        joinable = self.viper.And(joinable_func, post_pred, self._position,
                                  self._info)
        joinable = self.viper.Implies(tcond, joinable, self._position, self._info)
        return joinable

    def _remember_old_expressions(self, method: PythonMethod,
                                  collector: OldExpressionCollector) -> Expr:
        """
        Creates an assertion that connects all old expressions in the method's
        postcondition to the thread object.
        """
        old_info = self.viper.TrueLit(self._position, self._info)
        normalizer = OldExpressionTransformer()
        normalizer.arg_names = [arg for arg in method._args]
        for expr in collector.expressions:
            print_expr = normalizer.visit(copy.deepcopy(expr))
            name = pprint(print_expr)
            id = self.viper.IntLit(self._translator._get_string_value(name),
                                   self._position, self._info)
            old_func = self.viper.DomainFuncApp(GET_OLD_FUNC, [self._thread, id],
                                                self.viper.Ref, self._position,
                                                self._info, THREAD_DOMAIN)
            _, old_val = self._translator.translate_expr(expr, self._ctx)
            func_val = self.viper.EqCmp(old_func, old_val, self._position, self._info)
            old_info = self.viper.And(old_info, func_val, self._position, self._info)
        return old_info

    def _set_parameter_aliases(self, method: PythonMethod) -> List[Stmt]:
        """
        Sets var aliases in the context so that references to the parameters of this
        method are translated to references to the respective getArg calls instead.
        """
        arg_vars = []
        stmts = []
        for index, arg in enumerate(method._args.values()):
            arg_var = self._ctx.actual_function.create_variable(
                'thread_arg', arg.type, self._translator.translator)
            arg_vars.append(arg_var)
            index_lit = self.viper.IntLit(index, self._position, self._info)
            arg_expr = self.viper.DomainFuncApp(GET_ARG_FUNC,
                                                [self._thread, index_lit],
                                                self.viper.Ref, self._position,
                                                self._info, THREAD_DOMAIN)
            stmts.append(self.viper.LocalVarAssign(arg_var.ref(),
                                                   arg_expr,
                                                   self._position, self._info))
            type_info = self._translator.type_check(arg_var.ref(),
                                                    arg.type, self._position,
                                                    self._ctx)
            stmts.append(self.viper.Inhale(type_info, self._position, self._info))
        for index, name in enumerate(method._args):
            self._ctx.set_alias(name, arg_vars[index])
        return stmts

    def _add_precondition_exhales(self) -> None:
        """
        Creates statements that conditionally exhale the preconditions of the possible
        thread targets, inhale join permissions, remember values of old expressions in
        method postconditions.
        """
        method_id_type = self.viper.DomainType(METHOD_ID_DOMAIN, {}, [])
        actual_method = self.viper.DomainFuncApp(GET_METHOD_FUNC, [self._thread],
                                                 method_id_type, self._position,
                                                 self._info, THREAD_DOMAIN)
        for method in self._targets:
            # Forked methods must not have obligations in their postconditions.
            for instances in method.obligation_info._postcondition_instances.values():
                if instances:
                    raise InvalidProgramException(self._target_node,
                                                  'invalid.thread.start')
            collector = OldExpressionCollector()
            for post, _ in method.postcondition:
                collector.visit(post)

            stmts = []

            # To prepare translating the thread precondition, create alias variables for
            # each parameter and assign the argument values.
            arg_stmts = self._set_parameter_aliases(method)
            stmts.extend(arg_stmts)

            self._ctx.inlined_calls.append(method)
            old_loop_stack = self._ctx.obligation_context._loop_stack
            self._ctx.obligation_context._loop_stack = []
            self._ctx.ignore_waitlevel_constraints = True

            # Remember values of old expressions in postcondition.
            old_info = self._remember_old_expressions(method, collector)
            stmts.append(self.viper.Inhale(old_info, self._position, self._info))

            # Translate the actual precondition.
            pre_assertion = self.viper.TrueLit(self._position, self._info)
            for pre, _ in method.precondition:
                _, pre_val = self._translator.translate_expr(pre, self._ctx, impure=True,
                                                             target_type=self.viper.Bool)
                pre_assertion = self.viper.And(pre_assertion, pre_val, self._position,
                                               self._info)
            stmts.append(self.viper.Exhale(pre_assertion.whenExhaling(), self._position,
                                           self._info))

            # Inhale join permission.
            joinable = self._create_join_permission(method)
            stmts.append(self.viper.Inhale(joinable, self._position, self._info))

            for name in method._args:
                self._ctx.remove_alias(name)
            self._ctx.inlined_calls.pop()
            self._ctx.obligation_context._loop_stack = old_loop_stack
            self._ctx.ignore_waitlevel_constraints = False

            # Do all this under the condition that the current method is the thread's
            # actual target method.
            this_method = self.viper.DomainFuncApp(method.threading_id, [],
                                                   method_id_type, self._position,
                                                   self._info, METHOD_ID_DOMAIN)
            method_cond = self.viper.EqCmp(actual_method, this_method, self._position,
                                           self._info)
            then_block = self._translator.translate_block(stmts, self._position,
                                                          self._info)
            else_block = self._translator.translate_block([], self._position, self._info)
            conditional = self.viper.If(method_cond, then_block, else_block,
                                        self._position, self._info)
            self._statements.append(conditional)

    def _add_waitlevel(self) -> None:
        level = self.create_level_call(sil.RefExpr(self._thread))
        comp = self._create_level_below(level, self._ctx)
        comp = comp.translate(self._translator, self._ctx, self._position, self._info)
        self._statements.append(self.viper.Inhale(comp, self._position, self._info))

    def _create_level_below(
            self, expr: sil.PermExpression,
            ctx: Context) -> sil.BoolExpression:
        residue_level_var = sil.PermVar(ctx.actual_function.obligation_info.residue_level)
        obligation = self._obligation_manager.must_release_obligation
        fields = obligation.create_fields_untranslated()
        var = ctx.current_function.create_variable(
            '_r', ctx.module.global_module.classes[OBJECT_TYPE],
            self._translator.translator, local=False)
        op = operator.lt(self.create_level_call(sil.RefVar(var)), expr)
        for_perms = [sil.ForPerm(var.sil_name, f, op) for f in fields]
        return sil.BigAnd(for_perms + [operator.lt(residue_level_var, expr)])

    def _translate_level(self, node: ast.Call) -> sil.PermExpression:
        """Translate a call to ``Level``."""
        assert len(node.args) == 1
        arg = sil.RefVar(node.args[0])
        return self.create_level_call(arg)

    def create_level_call(self, expr: sil.RefExpression) -> sil.PermExpression:
        return sil.PermCall('Level', [sil.CallArg('r', sil.REF, expr)])

    def get_statements(self) -> List[Stmt]:
        return self._statements



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

import ast
import logging
import os
import nagini_contracts.io_builtins
import nagini_contracts.lock
import tokenize

from collections import OrderedDict
from nagini_contracts.contracts import CONTRACT_FUNCS, CONTRACT_WRAPPER_FUNCS
from nagini_contracts.io_contracts import IO_OPERATION_PROPERTY_FUNCS
from nagini_translation.analyzer_io import IOOperationAnalyzer
from nagini_translation.external.ast_util import mark_text_ranges
from nagini_translation.lib.constants import (
    CALLABLE_TYPE,
    IGNORED_IMPORTS,
    INT_TYPE,
    LEGAL_MAGIC_METHODS,
    LITERALS,
    MYPY_SUPERCLASSES,
    OBJECT_TYPE,
    TUPLE_TYPE,
)
from nagini_translation.lib.program_nodes import (
    ContainerInterface,
    GenericType,
    MethodType,
    OptionalType,
    ProgramNodeFactory,
    PythonClass,
    PythonExceptionHandler,
    PythonField,
    PythonGlobalVar,
    PythonIOOperation,
    PythonModule,
    PythonNode,
    PythonTryBlock,
    PythonType,
    PythonVar,
    PythonMethod,
    TypeVar,
    UnionType,
)
from nagini_translation.lib.resolver import get_target as do_get_target
from nagini_translation.lib.typedefs import Expr
from nagini_translation.lib.typeinfo import TypeInfo
from nagini_translation.lib.util import (
    construct_lambda_prefix,
    contains_stmt,
    get_func_name,
    get_parent_of_type,
    InvalidProgramException,
    is_io_existential,
    UnsupportedException,
)
from nagini_translation.lib.views import PythonModuleView
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union


logger = logging.getLogger('nagini_translation.analyzer')


class Analyzer(ast.NodeVisitor):
    """
    Walks through the Python AST and collects the structures to be translated.
    """

    def __init__(self, types: TypeInfo, path: str, selected: Set[str]):
        self._node_factory = None
        self.types = types
        self.global_module = PythonModule(types, self.node_factory, None, None, None)
        self.global_module.global_module = self.global_module
        file = path.split(os.sep)[-1]
        self.module = PythonModule(types, self.node_factory, '__main__',
                                   self.global_module, None,
                                   sil_names=self.global_module.sil_names, file=file)
        self.current_class = None
        self.current_function = None
        self.current_scopes = []
        self.contract_only = False
        self.module_paths = [os.path.abspath(path)]
        self.modules = OrderedDict()
        self.modules[os.path.abspath(path)] = self.module
        self.asts = {}
        self.io_operation_analyzer = IOOperationAnalyzer(
            self, self.node_factory)
        # Are we defining an IOExists block?
        self._is_io_existential = False
        self._aliases = {}  # Dict[str, PythonBaseVar]
        self.current_loop_invariant = None
        self.selected = selected
        self.deferred_tasks = []

    @property
    def stmt_container(self):
        if self.current_function is not None:
            return self.current_function
        else:
            return self.module

    @property
    def node_factory(self):
        if not self._node_factory:
            self._node_factory = ProgramNodeFactory()
        return self._node_factory

    def define_new(self, container: Union[PythonModule, PythonClass],
                   name: str, node: ast.AST) -> None:
        """
        Called when a new top level element named ``name`` is created in
        ``container``. Checks there is any existing element with the
        same name, and raises an exception in that case.
        """
        if isinstance(container, PythonModule):
            if name in container.classes:
                cls = container.classes[name]
                if cls.defined:
                    raise InvalidProgramException(node, 'multiple.definitions')
            if (name in container.global_vars and
                    hasattr(container.global_vars[name], 'value') and
                    isinstance(node, (ast.FunctionDef, ast.ClassDef))):
                raise InvalidProgramException(node, 'multiple.definitions')
        if (name in container.functions or
                name in container.methods or
                name in container.predicates or
                (isinstance(container, PythonClass) and
                 name in container.static_methods)):
            raise InvalidProgramException(node, 'multiple.definitions')

    def collect_imports(self, abs_path: str) -> None:
        """
        Parses the file at the given location, puts the result into self.asts.
        Scans the parsed file for Import-statements and adds all imported paths
        to self.modules.
        """
        if abs_path.startswith('mod$'):
            # This is a module that corresponds to a directory, so it has no
            # contents of its own.
            return
        with tokenize.open(abs_path) as file:
            text = file.read()
        parse_result = ast.parse(text)
        try:
            mark_text_ranges(parse_result, text)
        except Exception:
            # Ignore
            pass
        self.modules[abs_path].node = parse_result
        self.asts[abs_path] = parse_result
        assert isinstance(parse_result, ast.Module)
        for stmt in parse_result.body:
            if isinstance(stmt, ast.Import):
                for name in stmt.names:
                    module_name = name.name
                    if module_name in IGNORED_IMPORTS:
                        continue
                    redefined_name = name.asname
                    if not redefined_name:
                        redefined_name = module_name
                    assert module_name in self.types.files
                    path = os.path.abspath(self.types.files[module_name])
                    self.add_module(path, abs_path, redefined_name, parse_result)
            elif isinstance(stmt, ast.ImportFrom):
                module_name = stmt.module
                if module_name in IGNORED_IMPORTS:
                    continue
                if module_name == 'nagini_contracts.io_builtins':
                    path = nagini_contracts.io_builtins.__file__
                elif module_name == 'nagini_contracts.lock':
                    path = nagini_contracts.lock.__file__
                else:
                    assert module_name in self.types.files
                    path = self.types.files[module_name]
                path = os.path.abspath(path)
                if len(stmt.names) == 1 and stmt.names[0].name == '*':
                    names = None
                else:
                    names = [(name.name, name.asname if name.asname else None)
                             for name in stmt.names] # TODO rename?
                self.add_module(path, abs_path, None, parse_result, names)
        self.module_index = self.module_paths.index(abs_path) + 1

    def add_module(self, abs_path: str, into: str, as_name: Optional[str],
                   node: ast.Module, names: List[Tuple[str, str]] = None) -> None:
        """
        Adds the module with the given 'abs_path' into the module with
        path 'into'. If it's a from-import, 'as_name' should be None and
        'names' may contain rename information for specific imported
        members, otherwise (with a normal import) as_name should contain the
        name under which the module is imported.
        """
        if as_name and '.' in as_name:
            # Imported multi.part.name, add part by part
            split = as_name.split('.', 1)
            first = split[0]
            rest = split[1]
            # Add first part into original module. The first part must
            # correspond to a directory, so it has no contents of its own,
            # and we just create a dummy module (indicated by the path starting
            # with 'mod$') that's just there to contain the following module(s).
            self.add_module('mod$' + first, into, first, node)
            # Add next part(s) into first part recursively
            self.add_module(abs_path, 'mod$' + first, rest, node)
            return
        if abs_path not in self.module_paths:
            # Module has not been imported yet
            self.module_paths.insert(self.module_index, abs_path)
            type_prefix = self.types.get_type_prefix(abs_path)
            if type_prefix:
                # Get the part of the path that corresponds to modules (i.e.
                # drop prefix that just goes to library directory).
                file = abs_path[abs_path.index(type_prefix.split('.')[0]):]
            else:
                file = None
            new_module = PythonModule(self.module.types, self.node_factory,
                                      type_prefix, self.module.global_module, node,
                                      self.module.sil_names, file)
            self.modules[abs_path] = new_module
            self.collect_imports(abs_path)
        else:
            new_module = self.modules[abs_path]
        into_mod = self.modules[into]
        if as_name:
            assert not names
            into_mod.namespaces[as_name] = new_module
        else:
            if names:
                new_module = PythonModuleView(new_module, names)
            into_mod.from_imports.append(new_module)


    def analyze(self) -> None:
        """
        Recursively analyzes the main module and all imported modules.
        """
        self.visited_modules = [self.module]
        self.visit_module(self.module)

    def process(self, translator: 'Translator') -> None:
        """
        Performs preprocessing on the result of the analysis, which infers some
        things, creates some data structures for the translation etc.
        """
        self.module.global_module.process(translator)
        for module in self.modules.values():
            module.process(translator)

    def add_native_silver_builtins(self, interface: Dict) -> None:
        """
        Adds the classes, methods and functions in the interface-dict to
        the program. Meant to be used with a dict containing all methods/...
        that have native Silver representations and won't be created by the
        translator.
        """
        # Create global classes first
        for class_name in interface:
            if class_name == 'global':
                continue
            cls = self.find_or_create_class(class_name,
                                            module=self.module.global_module)
            cls.interface = True
            cls.defined = True
        for class_name in interface:
            self._process_interface_class(class_name, interface[class_name])

    def _process_interface_class(self, class_name: str, if_cls: Dict,
                                 node_factory: ProgramNodeFactory = None):
        if class_name == 'global':
            cls = None
        else:
            cls = self.find_or_create_class(class_name)
            object_class = self.find_or_create_class(OBJECT_TYPE,
                                                     self.module.global_module)
            if 'type_vars' in if_cls:
                for i in range(if_cls['type_vars']):
                    name = 'var' + str(i)
                    cls.type_vars[name] = TypeVar(
                        name, object_class, self.module.global_module,
                        target_type=cls, index=i)
            if 'extends' in if_cls:
                superclass = self.find_or_create_class(
                    if_cls['extends'], module=self.module.global_module)
                cls.superclass = superclass

        for method_name in if_cls.get('methods', []):
            if_method = if_cls['methods'][method_name]
            self._add_native_silver_method(
                method_name, if_method, cls, False, node_factory=node_factory)
        for method_name in if_cls.get('functions', []):
            if_method = if_cls['functions'][method_name]
            self._add_native_silver_method(
                method_name, if_method, cls, True, node_factory=node_factory)
        for pred_name in if_cls.get('predicates', []):
            if_pred = if_cls['predicates'][pred_name]
            self._add_native_silver_method(
                pred_name, if_pred, cls, True, True, node_factory=node_factory)

    def _add_native_silver_method(
            self, method_name: str, if_method: Dict[str, Any], cls: PythonClass,
            pure: bool, predicate: bool = False,
            node_factory: ProgramNodeFactory = None) -> None:
        if not node_factory:
            node_factory = self.node_factory
        method = node_factory.create_python_method(
            method_name, None, cls, self.module.global_module, pure, False, node_factory,
            interface=True, interface_dict=if_method)
        if 'display_name' in if_method:
            method.interface_name = if_method['display_name']
        ctr = 0
        for arg_type in if_method['args']:
            name = 'arg_' + str(ctr)
            arg = node_factory.create_python_var(
                name, None, self.find_or_create_class(arg_type,
                                                      self.module.global_module))
            ctr += 1
            method.add_arg(name, arg)
        if if_method['type']:
            method.type = self.find_or_create_class(if_method['type'],
                                                    self.module.global_module)
        if if_method.get('generic_type'):
            method.generic_type = if_method['generic_type']
        if if_method.get('requires'):
            method.requires = if_method['requires']
        cont = cls if cls else self.module.global_module
        if predicate:
            cont.predicates[method_name] = method
        elif pure:
            cont.functions[method_name] = method
        else:
            cont.methods[method_name] = method

    def visit_module(self, module: PythonModule) -> None:
        self.visit(module.node, None)

    def visit_Module(self, node: ast.Module) -> None:
        for stmt in node.body:
            if isinstance(stmt, (ast.ClassDef, ast.FunctionDef, ast.Import,
                                 ast.ImportFrom, ast.Assign)):
                continue
            if (isinstance(stmt, ast.Expr) and
                    isinstance(stmt.value, ast.Str)):
                # A docstring.
                continue
            if get_func_name(stmt) == 'Import':
                continue
            if get_func_name(stmt) in CONTRACT_WRAPPER_FUNCS:
                raise InvalidProgramException(stmt, 'invalid.contract.position')
        self.visit_default(node)

    def visit_default(self, node: ast.AST) -> None:
        for field in node._fields:
            fieldval = getattr(node, field)
            if isinstance(fieldval, ast.AST):
                self.visit(fieldval, node)
            elif isinstance(fieldval, list):
                for item in fieldval:
                    if isinstance(item, ast.AST):
                        self.visit(item, node)

    def visit(self, child_node: ast.AST, parent: ast.AST) -> None:
        child_node._parent = parent
        method = 'visit_' + child_node.__class__.__name__
        visitor = getattr(self, method, self.visit_default)
        visitor(child_node)

    def visit_but_ignore(self, node: ast.AST, parent: ast.AST) -> None:
        node._parent = parent
        for field in node._fields:
            fieldval = getattr(node, field)
            if isinstance(fieldval, ast.AST):
                self.visit_but_ignore(fieldval, node)
            elif isinstance(fieldval, list):
                for item in fieldval:
                    if isinstance(item, ast.AST):
                        self.visit_but_ignore(item, node)

    def get_target(self, node: ast.AST,
                   container: ContainerInterface) -> PythonNode:
        """
        Finds the PythonNode that the given 'node' refers to, e.g. a PythonClass
        or a PythonVar, if the immediate container (e.g. a PythonMethod) of the
        node is 'container'.
        """
        containers = [container]
        if isinstance(container, (PythonMethod, PythonIOOperation)):
            containers.extend(container.module.get_included_modules())
        else:
            containers.extend(container.get_included_modules(()))
        return do_get_target(node, containers, container)

    def find_or_create_class(self, name: str, module=None) -> PythonClass:
        """
        Gets the class with the given 'name' from the given 'module' (or the
        current module if none is provided). If no such class exists, one will
        be created.
        """
        # If the name refers to a type alias from the typing module, get
        # the actual name
        aliases = {'List': 'list',
                   'Tuple': 'tuple',
                   'Set': 'set',
                   'Dict': 'dict',
                   'Type': 'type',}
        name = aliases.get(name, name)
        if self.current_class and name in self.current_class.type_vars:
            return self.current_class.type_vars[name]
        if not module:
            module = self.module
        # Check all imported modules for the class.
        for visible_module in module.get_included_modules((), True):
            if name in visible_module.classes:
                cls = visible_module.classes[name]
                break
        else:
            # Class doesn't exist yet, create it.
            cls = self.node_factory.create_python_class(name, module,
                                                        self.node_factory)
            module.classes[name] = cls
        return cls

    def find_or_create_target_class(self, node: ast.AST) -> PythonClass:
        """
        Assuming that the given node is a reference to a class, retrieves the
        PythonClass object if it already exists, otherwise creates one.
        """
        if isinstance(node, ast.Name):
            return self.find_or_create_class(node.id)
        elif isinstance(node, ast.Str):
            return self.find_or_create_class(node.s)
        elif isinstance(node, ast.Attribute):
            ctx = self.get_target(node.value, self.module)
            return self.find_or_create_class(node.attr, module=ctx)
        elif isinstance(node, ast.Subscript):
            cls = self.find_or_create_target_class(node.value)
            if isinstance(node.slice.value, ast.Name):
                ast_args = [node.slice.value]
            elif isinstance(node.slice.value, ast.Str):
                ast_args = [node.slice.value]
            else:
                ast_args = node.slice.value.elts
            args = [self.find_or_create_target_class(arg) for arg in ast_args]
            result = GenericType(cls, args)
            return result
        else:
            raise UnsupportedException(node,
                                       'class literal has unsupported format')

    def _check_is_adt_well_formed(self, adt: PythonClass,
                                  actual_bases: List[ast.AST],
                                  node: ast.AST, ast: ast.AST):
        """
        Checks if the ADT is syntactically well-formed, throwing an
        invalid program exception otherwise.
        """
        # ADT classes can have at most two superclasses
        if len(actual_bases) > 2:
            raise InvalidProgramException(adt.node, 'malformed.adt',
                    'malformed algebraic data type: superclasses can only ' +
                    'be a class optionally followed by one NamedTuple')
        # ADT classes should have no body
        if not (len(node.body) == 1 and isinstance(node.body[0], ast.Pass) or
                len(node.body) == 2 and isinstance(node.body[0], ast.Expr)
                                    and isinstance(node.body[1], ast.Pass)):
            raise InvalidProgramException(adt.node, 'malformed.adt',
                    'malformed algebraic data type: classes cannot have ' +
                    'body, fields should be defined in NamedTuples instead')
        elif len(actual_bases) == 2:
            # ADT classes can only be defined via NamedTuple
            if not (isinstance(actual_bases[1], ast.Call)
                and actual_bases[1].func.id == 'NamedTuple'):
                raise InvalidProgramException(actual_bases[1], 'malformed.adt',
                    'malformed algebraic data type: only NamedTuple can be ' +
                    'used to define constructors')
            # The name of the NamedTuple should match the ADT constructor being
            # defined
            if not (adt.name == actual_bases[1].args[0].s):
                raise InvalidProgramException(actual_bases[1], 'malformed.adt',
                    'malformed algebraic data type: name of NamedTuple has to ' +
                    'be the same of the class (ADT Constructor)')
            # ADT constructor cannot be a direct subclass of 'ADT' since there
            # should be an intermediary class in the hierarchy defining the ADT
            # class the ADT constructor should be part of (i.e. subclass of)
            if adt.superclass.name == 'ADT':
                raise InvalidProgramException(node, 'malformed.adt',
                    'malformed algebraic datatype: ADT constructor classes ' +
                    'cannot inherit directly from ADT class, instead they ' +
                    'should inherit from their own ADT defining class, which ' +
                    'should be a direct subclass of ADT')
        else: # len(actual_bases) == 1
            # ADT defining classes should inherit directly from ADT class
            if adt.superclass.name != 'ADT':
                raise InvalidProgramException(node, 'malformed.adt',
                    'malformed algebraic datatype: ADT defining classes ' +
                    'should be direct subclasses of ADT class')

    def _visit_ADT(self, cls: PythonClass, actual_bases: List[ast.AST],
                   node: ast.AST, ast: ast.AST) -> List[ast.AST]:
        """
        Parses an ADT if well-formed, throwing an exception otherwise. The
        ADT is defined by a class hierarchy (sum) and a NamedTuple (product).
        """
        self._check_is_adt_well_formed(cls, actual_bases, node, ast)
        cls.is_adt = True
        if len(actual_bases) == 2:
            # Parse NamedTuple's fields and their respective types
            for field_decl in actual_bases[1].args[1].elts:
                field_name, field_type = field_decl.elts
                cls.add_field(field_name.s, field_decl,
                    self.get_target(field_type, self.module).try_unbox())
            # Consider ADTs as a special case of single inheritance
            actual_bases = [actual_bases[0]]
        return actual_bases

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        if self.current_function or self.current_class:
            raise InvalidProgramException(node, 'nested.class.declaration')
        name = node.name
        self.define_new(self.module, name, node)
        cls = self.find_or_create_class(name)
        cls.defined = True
        cls.node = node
        self.current_class = cls
        actual_bases = []
        current_index = 0
        for base in node.bases:
            if isinstance(base, ast.Subscript) and base.value.id == 'Generic':
                if isinstance(base.slice.value, ast.Name):
                    arg_names = [base.slice.value.id]
                else:
                    arg_names = [elmt.id for elmt in base.slice.value.elts]
                for arg_name in arg_names:
                    assert arg_name in self.module.type_vars
                    var_info = self.module.type_vars[arg_name]
                    bound = self.convert_type(var_info[0], base)
                    var = TypeVar(arg_name, bound, self.module,
                                  target_type=cls, index=current_index)
                    cls.type_vars[arg_name] = var
                    current_index += 1
            else:
                # Drop stuff like Sized that is only needed for mypy and does
                # not correspond to an actual superclass.
                if not (isinstance(base, ast.Name) and
                        base.id in MYPY_SUPERCLASSES):
                    actual_bases.append(base)

        if len(actual_bases) > 0:
            cls.superclass = self.find_or_create_target_class(actual_bases[0])
            if isinstance(cls.superclass, PythonClass) and cls.superclass.is_adt:
                actual_bases = self._visit_ADT(cls, actual_bases, node, ast)
        if len(actual_bases) > 1:
            raise UnsupportedException(node, 'multiple inheritance')
        if len(actual_bases) == 0:
            cls.superclass = self.find_or_create_class(OBJECT_TYPE)
        if cls.python_class not in cls.superclass.python_class.direct_subclasses:
            cls.superclass.python_class.direct_subclasses.append(cls.python_class)

        for member in node.body:
            self.visit(member, node)
        self.current_class = None

    def _is_illegal_magic_method_name(self, name: str) -> bool:
        """
        Anything that could potentially be a magic method, i.e. anything that
        has __this__ form, is considered illegal unless it is one of the names
        we explicitly support.
        """
        if name.startswith('__') and name.endswith('__'):
            if name not in LEGAL_MAGIC_METHODS:
                return True
        return False

    def analyze_import(self, module_name: str) -> None:
        """
        Visits the given module if it hasn't already been visited.
        """
        if module_name in IGNORED_IMPORTS:
            return
        for mod in self.modules.values():
            if mod.type_prefix == module_name:
                module = mod
                break
            if mod.type_prefix == '__main__':
                main_module = mod
        else:
            # type prefix of the main module is __main__, therefore isn't found.
            module = main_module

        if module in self.visited_modules:
            return
        self.visited_modules.append(module)
        old_contract_only = self.contract_only
        self.contract_only = True
        old_module = self.module
        self.module = module
        self.visit_module(module)
        self.module = old_module
        self.contract_only = old_contract_only

    def visit_Import(self, node: ast.Import):
        for name in node.names:
            self.analyze_import(name.name)

    def visit_ImportFrom(self, node: ast.ImportFrom):
        self.analyze_import(node.module)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        if self.current_function:
            raise InvalidProgramException(node, 'nested.function.declaration')
        name = node.name
        if self._is_illegal_magic_method_name(name):
            raise InvalidProgramException(node, 'illegal.magic.method')
        assert isinstance(name, str)
        if self.is_io_operation(node):
            self.io_operation_analyzer.analyze_io_operation(node)
            return
        if self.current_class is None:
            scope_container = self.module
        else:
            scope_container = self.current_class
        self.define_new(scope_container, name, node)
        is_property_getter = self.is_property_getter(node)
        is_property_setter = self.is_property_setter(node)
        if is_property_getter or is_property_setter:
            container = scope_container.fields
        elif self.is_predicate(node):
            container = scope_container.predicates
        elif self.is_pure(node):
            container = scope_container.functions
        elif self.is_static_method(node):
            container = scope_container.static_methods
        else:
            container = scope_container.methods
        if not (is_property_getter or is_property_setter) and name in container:
            func = container[name]
            if not self.is_static_method(node):
                func.cls = self.current_class
            func.pure = self.is_pure(node)
            func.node = node
            func.superscope = scope_container
        else:
            pure = is_property_getter or self.is_pure(node)
            if pure:
                contract_only = self.is_declared_contract_only(node)
            else:
                contract_only = self.contract_only or self.is_contract_only(node)
            func = self.node_factory.create_python_method(
                name, node, self.current_class, scope_container, pure,
                contract_only, self.node_factory)
            if is_property_setter:
                container[name].setter = func
            else:
                container[name] = func
        if self.is_static_method(node):
            func.method_type = MethodType.static_method
        elif self.is_class_method(node):
            func.method_type = MethodType.class_method
            self.current_class._has_classmethod = True
        func.predicate = self.is_predicate(node)
        func.all_low = self.is_all_low(node)
        func.preserves_low = self.preserves_low(node)

        # TODO: When we want to support method type parameters, this would be
        # the place to find all type variables used in the parameters which
        # don't stem from the class and add them to func.type_vars.
        functype = self.module.get_func_type(func.scope_prefix)
        if func.pure and not functype:
            raise InvalidProgramException(node, 'function.type.none')
        self.current_function = func

        self.visit(node.args, node)

        if not is_property_setter:
            func.type = self.convert_type(functype, node)

        for child in node.body:
            if is_io_existential(child):
                self._is_io_existential = True
                self.visit(child.value.args[0], node)
            else:
                self.visit(child, node)
        self.current_function = None

    def visit_loop(self, node: Union[ast.While, ast.For]) -> None:
        old_loop_invariant = self.current_loop_invariant
        self.stmt_container.loop_invariants[node] = []
        self.current_loop_invariant = self.stmt_container.loop_invariants[
            node]
        if isinstance(node, ast.While):
            self.visit(node.test, node)
        else:
            self.visit(node.target, node)
        for child in node.body:
            if is_io_existential(child):
                self._is_io_existential = True
                self.visit(child.value.args[0], node)
            else:
                self.visit(child, node)
        if node.orelse:
            for stmt in node.orelse:
                self.visit(stmt, node)
        self.current_loop_invariant = old_loop_invariant

    def visit_While(self, node: ast.While) -> None:
        self.visit_loop(node)

    def visit_For(self, node: ast.While) -> None:
        self.visit_loop(node)

    def visit_Assign(self, node: ast.Assign) -> None:
        is_alias = False
        is_type_var = False
        # Check if this is a type alias
        if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
            actual_prefix = []
            if self.module.type_prefix:
                actual_prefix = self.module.type_prefix.split('.')
            actual_prefix.append(node.targets[0].id)
            lhs_name = tuple(actual_prefix)
            # If it's a type alias marked by mypy
            if lhs_name in self.types.type_aliases:
                type_name = self.types.type_aliases[lhs_name]
                aliased_type = self.convert_type(type_name, node)
                self.module.classes[node.targets[0].id] = aliased_type
                is_alias = True
            # If it's a type variable marked by mypy
            elif lhs_name in self.types.type_vars:
                var = self.types.type_vars[lhs_name]
                self.module.type_vars[node.targets[0].id] = var
                is_type_var = True
            # Could still be a type alias if RHS refers to class
            elif not isinstance(node.value, ast.Call):
                target = self.get_target(node.value, self.module)
                if isinstance(target, PythonType):
                    self.module.classes[node.targets[0].id] = target
                    is_alias = True

        # Nothing else to do for type aliases and type vars, for all other
        # cases proceed as usual.
        if is_alias:
            if self.current_function or self.current_class:
                raise InvalidProgramException(node, 'local.type.alias')
        elif is_type_var:
            if self.current_function or self.current_class:
                raise InvalidProgramException(node, 'local.typevar')
        else:
            self.visit_default(node)

    def visit_arguments(self, node: ast.arguments) -> None:
        assert self.current_function is not None
        for arg in node.args:
            self.visit(arg, node)
        self.current_function.nargs = len(node.args)
        for kw_only in node.kwonlyargs:
            self.visit(kw_only, node)
        defaults = node.defaults
        args = list(self.current_function.args.values())
        for index in range(len(defaults)):
            arg = args[index - len(defaults)]
            arg.default = defaults[index]
        if node.vararg:
            arg = node.vararg
            annotated_type = self.typeof(arg)
            assert annotated_type.name == TUPLE_TYPE
            annotated_type.exact_length = False
            var_arg = self.node_factory.create_python_var(arg.arg, arg,
                                                          annotated_type)
            self.current_function.var_arg = var_arg
        if node.kwarg:
            arg = node.kwarg
            annotated_type = self.typeof(arg)
            kw_arg = self.node_factory.create_python_var(arg.arg, arg,
                                                         annotated_type)
            self.current_function.kw_arg = kw_arg

    def visit_ListComp(self, node: ast.Lambda) -> None:
        name = construct_lambda_prefix(node.lineno, node.col_offset)
        target = node.generators[0].target
        local_name = name + '$' + target.id
        var = self.node_factory.create_python_var(
            target.id, target, self.typeof(target))
        self.current_function.special_vars[local_name] = var
        self.visit_but_ignore(node, node._parent)
        return

    def visit_Lambda(self, node: ast.Lambda) -> None:
        assert self.current_function
        name = construct_lambda_prefix(node.lineno, node.col_offset)
        self.current_scopes.append(name)
        for arg in node.args.args:
            # For IO operation arguments, we want parameters to be primitive
            # b/c primitive equality is simpler than boxed reference equality,
            # and for simplicity we use the same policy for output params too.
            if self._is_io_existential:
                arg_type = self.typeof(arg).try_unbox()
                var = self.node_factory.create_python_io_existential_var(
                    arg.arg, arg, arg_type)
                self._aliases[arg.arg] = var
            else:
                var = self.node_factory.create_python_var(
                    arg.arg, arg, self.typeof(arg))
            alts = self.get_alt_types(node)
            var.alt_types = alts
            local_name = name + '$' + arg.arg
            if self._is_io_existential:
                self.current_function.io_existential_vars[local_name] = var
            else:
                self.current_function.special_vars[local_name] = var
        self._is_io_existential = False
        self.visit(node.body, node)
        self.current_scopes.pop()
        for arg in node.args.args:
            self._aliases.pop(arg.arg, None)

    def visit_arg(self, node: ast.arg) -> None:
        assert self.current_function is not None
        node_type = self.typeof(node)
        self.current_function.args[node.arg] = \
            self.node_factory.create_python_var(node.arg, node, node_type)
        # If we just introduced new type variables, create the expression that
        # represents their value.
        for type_var in self.current_function.type_vars.values():
            if type_var.type_expr:
                continue
            type_var.type_expr = self.create_type_var_definition(type_var,
                                                                 node_type)
        alts = self.get_alt_types(node)
        self.current_function.args[node.arg].alt_types = alts

    def create_type_var_definition(
            self, type_var: TypeVar, node_type: PythonType) -> Callable[
                ['Translator', Expr, 'Context'], Expr]:
        """
        Creates a function that returns a type expression which defines the
        given type var when given the type of the defining expression, which has
        type ``node_type``.
        """
        if isinstance(node_type, PythonClass):
            return None
        if isinstance(node_type, GenericType):
            for i, arg in enumerate(node_type.type_args):
                # Check if the type argument defines the type variable, i.e. if
                # the recursive call returns a definition.
                arg_def = self.create_type_var_definition(type_var, arg)
                if arg_def:
                    def result(translator, typ, ctx):
                        previous = arg_def(translator, typ, ctx)
                        return translator.get_type_arg(previous, node_type, i,
                                                       ctx)
                    return result
        if node_type is type_var:
            def result(translator, typ, ctx):
                return typ
            return result
        return None

    def track_access(self, node: ast.AST, var: Union[PythonVar, PythonField]) -> None:
        if var is None:
            return
        if isinstance(node.ctx, ast.Load):
            var.reads.append(node)
        elif isinstance(node.ctx, ast.Store):
            var.writes.append(node)

    def visit_Call(self, node: ast.Call) -> None:
        """
        Collects preconditions, postconditions, raised exceptions and
        invariants.
        """
        if (isinstance(node.func, ast.Name) and
                node.func.id in CONTRACT_WRAPPER_FUNCS):
            if node.func.id == 'Requires':
                self.stmt_container.precondition.append(
                    (node.args[0], self._aliases.copy()))
            elif node.func.id == 'Ensures':
                self.stmt_container.postcondition.append(
                    (node.args[0], self._aliases.copy()))
            elif node.func.id == 'Exsures':
                exception = self.get_target(node.args[0], self.module)
                if exception not in self.current_function.declared_exceptions:
                    self.current_function.declared_exceptions[exception] = []
                self.current_function.declared_exceptions[exception].append(
                    (node.args[1], self._aliases.copy()))
            elif node.func.id == 'Invariant':
                self.current_loop_invariant.append(
                    (node, self._aliases.copy()))
        if (isinstance(node.func, ast.Name) and
            node.func.id in IO_OPERATION_PROPERTY_FUNCS):
            raise InvalidProgramException(
                node, 'invalid.io_operation.misplaced_property')
        if isinstance(node.func, ast.Name) and node.func.id == 'Thread':
            return
        if isinstance(node.func, ast.Name) and node.func.id == 'getOld':
            return
        if isinstance(node.func, ast.Name) and node.func.id == 'LowEvent':
            preconditions = list(map(lambda tuple: tuple[0], self.stmt_container.precondition))
            if not contains_stmt(preconditions, node):
                raise InvalidProgramException(node, 'invalid.contract.position')
        self.visit_default(node)

    def _get_parent_of_type(self, node: ast.AST, typ: type) -> ast.AST:
        """
        Returns the closest parent node of 'node' that is of the given type
        (e.g. ast.Name), or None if there is no such node.
        """
        parent = node._parent
        while not isinstance(parent, ast.Module):
            if isinstance(parent, typ):
                return parent
            parent = parent._parent
        return None

    def _get_parents_of_type(self, node: ast.AST, typ: type) -> List[ast.AST]:
        """
        Returns all parent nodes of 'node' that are of the given type (e.g.
        ast.Name), sorted from bottom (closest to 'node') to the top of the
        tree. If there is no such parent node, the returned list will be empty.
        """
        result = []
        current = get_parent_of_type(node, typ)
        while current:
            result.append(current)
            current = get_parent_of_type(current, typ)
        return result

    def visit_Name(self, node: ast.Name) -> None:
        """
        This method tracks all local variables, global variables,
        lambda arguments and static fields in the program and creates the
        corresponding PythonNodes.
        """
        # Since lots of things in the AST (that we don't care about here) are
        # also ast.Names, we filter out those cases first.
        if node.id in LITERALS:
            return
        if node.id.startswith('IOExists'):
            raise InvalidProgramException(node, 'invalid.ioexists.misplaced')
        if get_parent_of_type(node, ast.Call):
            return
        arg_parent = get_parent_of_type(node, ast.arg)
        if arg_parent:
            if arg_parent.annotation is node:
                # We're looking at a parameter type annotation
                return
        func_def_parent = get_parent_of_type(node, ast.FunctionDef)
        if func_def_parent:
            if func_def_parent.returns is node:
                # We're looking at a return type annotation
                return
            if node in func_def_parent.decorator_list:
                # We're looking at a decorator
                return
        handler_parents = self._get_parents_of_type(node, ast.ExceptHandler)
        for handler_parent in handler_parents:
            if handler_parent:
                if handler_parent.type is node:
                    # We're looking at an exception type that's caught
                    return
                if handler_parent.name and handler_parent.name == node.id:
                    # We're looking at an alias for a caught exception,
                    # those are handled along with the try-block itself.
                    return
        # Node could be global reference or static member
        # or local variable or function argument.
        if self.current_function is None:
            # Node is global in some way.
            if self.current_class is None:
                if node.id in self.module.classes:
                    return
                # Node is a global variable.
                if isinstance(node.ctx, ast.Store):
                    cls = self.typeof(node)
                    self.define_new(self.module, node.id, node)
                    existing_var = self.get_target(node, self.module)
                    if existing_var:
                        var = existing_var
                    else:
                        var = self.node_factory.create_python_global_var(
                            node.id, node, cls, self.module)
                    assign = node._parent
                    if isinstance(assign, ast.Assign) and len(assign.targets) == 1:
                        var.value = assign.value
                    self.module.global_vars[node.id] = var
                current_module = self.module
                # track access later
                def todo():
                    var = self.get_target(node, current_module)
                    if isinstance(var, PythonGlobalVar):
                        self.track_access(node, var)
                self.deferred_tasks.append(todo)
                return
            else:
                # Node is a static field.
                if isinstance(node.ctx, ast.Load):
                    return
                cls = self.typeof(node)
                self.define_new(self.current_class, node.id, node)
                var = self.node_factory.create_static_field(node.id, node, cls,
                                                            self.module,
                                                            self.current_class)
                assign = node._parent
                if (not isinstance(assign, ast.Assign)
                        or len(assign.targets) != 1):
                    msg = ('only simple assignments and reads allowed for '
                           'static fields')
                    raise UnsupportedException(assign, msg)
                var.value = assign.value
                self.current_class.static_fields[node.id] = var
                if node.id in self.current_class.fields:
                    # It's possible that we encountered a read of this field
                    # before seeing the definition, assumed it's a normal
                    # (non-static) field, and created the field. We remove it
                    # again now that we now it's actually static.
                    del self.current_class.fields[node.id]
                return
        # We're in a function
        if isinstance(node.ctx, ast.Store):
            # Node is a local variable, lambda argument, or a global variable
            # that hasn't been encountered yet
            var = None
            # First check if the node refers to something local that we already
            # know; in that case we don't have to do anything
            if node.id in self.current_function.locals:
                var = self.current_function.locals[node.id]
            elif node.id in self.current_function.args:
                pass
            elif node.id in self.current_function.special_vars:
                pass
            elif node.id in self.current_function.io_existential_vars:
                pass
            elif (self.current_function.var_arg and
                    self.current_function.var_arg.name == node.id):
                pass
            elif (self.current_function.kw_arg and
                    self.current_function.kw_arg.name == node.id):
                pass
            elif isinstance(self.get_target(node, self.module), PythonModule):
                return
            elif isinstance(self.get_target(node, self.module), PythonClass):
                return
            else:
                # We don't know this identifier yet, so it must be something
                # new.
                if node.id not in self.current_function.globals:
                    # Assume it's the first write to a local variable
                    var = self.node_factory.create_python_var(node.id,
                                                              node,
                                                              self.typeof(node))
                    alts = self.get_alt_types(node)
                    var.alt_types = alts
                    self.current_function.locals[node.id] = var
                else:
                    var = self.get_target(node, self.module)
                    if not isinstance(var, PythonGlobalVar):
                        # Assume it's the first write to a local variable
                        var = self.node_factory.create_python_global_var(
                            node.id, node, self.typeof(node), self.module)
                        alts = self.get_alt_types(node)
                        var.alt_types = alts
                        self.module.global_vars[node.id] = var
                    self.track_access(node, var)


            self.track_access(node, var)

    def visit_Attribute(self, node: ast.Attribute) -> None:
        """
        Tracks field accesses to find out which fields exist.
        """
        self.visit_default(node)
        # If the LHS of this attribute is a module, or if this
        # is part of a complex expression and it's guaranteed that the field
        # is also accessed in a simpler expression elsewhere, we just do
        # nothing here.
        if (not isinstance(node._parent, ast.Call) and
                not isinstance(node.value, ast.Subscript) and
                not (isinstance(node.value, ast.Call) and
                         isinstance(node.ctx, ast.Load))):
            container = self.current_function if self.current_function else self.module
            target = self.get_target(node.value, container)
            if isinstance(target, (PythonModule, PythonClass)):
                real_target = self.get_target(node, self.module)
                if isinstance(real_target, PythonGlobalVar):
                    self.track_access(node, real_target)
            else:
                receiver = self.typeof(node.value)
                if (isinstance(receiver, UnionType) and
                        not isinstance(receiver, OptionalType)):
                    for type in receiver.get_types() - {None}:
                        field = type.add_field(node.attr, node, self.typeof(node))
                        if isinstance(field, PythonField):
                            self.track_access(node, field)
                else:
                    field = receiver.python_class.add_field(node.attr, node,
                                                            self.typeof(node))
                    if isinstance(field, PythonField):
                        self.track_access(node, field)

    def visit_Global(self, node: ast.Global) -> None:
        for name in node.names:
            if name in self.current_function.locals:
                raise InvalidProgramException('invalid.global.statement')
            self.current_function.globals.add(name)

    def visit_Nonlocal(self, node: ast.Nonlocal) -> None:
        raise UnsupportedException(node)

    def convert_type(self, mypy_type, node) -> PythonType:
        """
        Converts an internal mypy type to a PythonType.
        """
        if (self.types.is_void_type(mypy_type) or
                self.types.is_none_type(mypy_type)):
            result = None
        elif self.types.is_instance_type(mypy_type):
            result = self.convert_type(mypy_type.type, node)
            if mypy_type.args:
                args = [self.convert_type(arg, node) for arg in mypy_type.args]
                if mypy_type.type.name() == 'enumerate':
                    # We cheat and represent type enumerate as a list of pairs.
                    assert len(args) == 1
                    int_type = self.module.global_module.classes[INT_TYPE]
                    tuple_class = self.module.global_module.classes[TUPLE_TYPE]
                    arg = GenericType(tuple_class, [int_type, args[0]])
                    args = [arg]
                result = GenericType(result, args)
        elif self.types.is_normal_type(mypy_type):
            return self._convert_normal_type(mypy_type)
        elif self.types.is_tuple_type(mypy_type):
            if hasattr(mypy_type, 'fallback'):
                # Handle ADTs which are typed as tuple and should fallback to class
                try:
                    fallback_type = self.convert_type(mypy_type.fallback, node)
                    if isinstance(fallback_type, PythonClass) and fallback_type.is_adt:
                        return fallback_type
                except UnsupportedException:
                    # When attempting to fallback a tuple that is not an ADT
                    # an exception could be raised (for example trying to fallback
                    # to Any). In this case we capture the exception and proceed
                    # without falling back (it is definitely not an ADT).
                    pass
            # Regular tuple handling without falling back
            args = [self.convert_type(arg_type, node)
                    for arg_type in mypy_type.items]
            result = GenericType(self.module.global_module.classes[TUPLE_TYPE],
                                 args)
        elif self.types.is_union_type(mypy_type):
            return self._convert_union_type(mypy_type, node)
        elif self.types.is_type_var(mypy_type):
            return self._convert_type_var(mypy_type, node)
        elif self.types.is_type_type(mypy_type):
            return self._convert_type_type(mypy_type, node)
        elif self.types.is_callable_type(mypy_type):
            return self._convert_callable_type(mypy_type, node)
        else:
            msg = 'Unsupported type: {}'.format(mypy_type.__class__.__name__)
            raise UnsupportedException(node, desc=msg)
        return result

    def _convert_normal_type(self, mypy_type) -> PythonType:
        prefix = mypy_type._fullname
        name = mypy_type.name()
        if prefix == 'builtins.enumerate':
            # We cheat and represent type enumerate as a list of pairs.
            prefix = 'builtins.list'
            name = 'list'
        if prefix.endswith('.' + name):
            prefix = prefix[:-(len(name) + 1)]
        target_module = self.module
        for module in self.modules.values():
            if module.type_prefix == prefix:
                target_module = module
                break
        result = self.find_or_create_class(name,
                                           module=target_module)
        return result

    def _convert_callable_type(self, mypy_type, node) -> PythonType:
        return self.find_or_create_class(CALLABLE_TYPE, module=self.module.global_module)

    def _convert_union_type(self, mypy_type, node) -> PythonType:
        args = [self.convert_type(arg_type, node)
                for arg_type in mypy_type.items]
        optional = False
        if None in args:
            # It's an optional type, remember this and wrap it later
            optional = True
            args.remove(None)
        if len(args) > 1:
            result = UnionType(args)
        else:
            result = args[0]
        if optional:
            result = OptionalType(result)
        return result

    def _convert_type_var(self, mypy_type, node) -> PythonType:
        name = mypy_type.name
        assert name in self.module.type_vars
        if (self.current_class and name in self.current_class.type_vars):
            return self.current_class.type_vars[name]
        if name in self.current_function.type_vars:
            return self.current_function.type_vars[name]
        bound_type = self.convert_type(self.module.type_vars[name][0], node)
        type_var = TypeVar(name, bound_type, self.module, target_node=node)
        self.current_function.type_vars[name] = type_var
        return type_var

    def _convert_type_type(self, mypy_type, node) -> PythonType:
        name = 'type'
        type_class = self.module.global_module.classes['type']
        args = [self.convert_type(mypy_type.item, node)]
        return GenericType(type_class, args)

    def get_alt_types(self, node: ast.AST) -> Dict[int, PythonType]:
        """
        If the given node refers to a local variable, returns a dict with
        the alt-types of the given variable, i.e. the types it has at other
        places where it is referenced (e.g. because a reference happens after
        an isinstance-check).
        The format is a dict that maps line numbers to types; for anything but
        references to local vars, the dict is empty.
        """
        if isinstance(node, (ast.Name, ast.arg)):
            context = []
            if self.current_class is not None:
                context.append(self.current_class.name)
            if self.current_function is not None:
                context.append(self.current_function.name)
            name = node.id if isinstance(node, ast.Name) else node.arg
            _, alts = self.module.get_type(context, name)
            result = {}
            if alts:
                for line, type in alts.items():
                    result[line] = self.convert_type(type, node)
            return result
        else:
            return {}

    def typeof(self, node: ast.AST) -> PythonType:
        """
        Returns the type of the given AST node.
        """
        if isinstance(node, ast.Name):
            if node.id in LITERALS:
                raise UnsupportedException(node)
            if node.id in self.module.classes:
                return self.module.classes[node.id]
            context = []
            if self.current_class is not None:
                context.append(self.current_class.name)
            if self.current_function is not None:
                context.append(self.current_function.name)
            context.extend(self.current_scopes)
            type, alts = self.module.get_type(context, node.id)
            key = (node.lineno, node.col_offset)
            if alts and key in alts:
                return self.convert_type(alts[key], node)
            return self.convert_type(type, node)
        elif isinstance(node, ast.Attribute):
            receiver = self.typeof(node.value)
            if isinstance(receiver, UnionType) and not isinstance(receiver, OptionalType):
                set_of_types = set()
                for type_in_union in receiver.get_types() - {None}:
                    if isinstance(type_in_union, OptionalType):
                        context = [type_in_union.optional_type.name]
                    else:
                        context = [type_in_union.name]
                    type, _ = self.module.get_type(context, node.attr)
                    set_of_types.add(self.convert_type(type, node))
                return UnionType(list(set_of_types)) if len(set_of_types) > 1 else set_of_types.pop()
            if isinstance(receiver, OptionalType):
                context = [receiver.optional_type.name]
            else:
                context = [receiver.name]
            type, _ = self.module.get_type(context, node.attr)
            return self.convert_type(type, node)
        elif isinstance(node, ast.arg):
            # Special case for cls parameter of classmethods; for those, we
            # return the type 'type[C]', where C is the class the method
            # belongs to.
            if self.current_function.method_type == MethodType.class_method:
                args_list = self.current_function.node.args.args
                if args_list and node is args_list[0]:
                    cls = self.module.global_module.classes['type']
                    return GenericType(cls, [self.current_class])
            context = []
            if self.current_class is not None:
                context.append(self.current_class.name)
            context.append(self.current_function.name)
            context.extend(self.current_scopes)
            type, _ = self.module.get_type(context, node.arg)
            return self.convert_type(type, node)
        elif (isinstance(node, ast.Call) and
              isinstance(node.func, ast.Name)):
            if node.func.id in CONTRACT_FUNCS:
                if node.func.id == 'Result':
                    return self.current_function.type
                else:
                    raise UnsupportedException(node)
            elif node.func.id == 'cast':
                return self.find_or_create_target_class(node.args[0])
            else:
                raise UnsupportedException(node)
        elif isinstance(node, ast.Call) and isinstance(node.func,
                                                       ast.Attribute):
            receiver = self.typeof(node.func.value)
            method = receiver.get_func_or_method(node.func.attr)
            return method.type
        else:
            raise UnsupportedException(node)

    def _get_basic_name(self, node: Union[ast.Name, ast.Attribute]) -> str:
        """
        For a reference.of.this.kind, returns the last part ('kind' in this
        case).
        """
        if isinstance(node, ast.Name):
            return node.id
        else:
            return self._get_basic_name(node.value)

    def visit_With(self, node: ast.With) -> None:
        """
        With-blocks get translated to try-finally-blocks, so we create a
        PythonTryBlock here.
        """
        self.visit_default(node)
        try_name = self.stmt_container.get_fresh_name('with')
        try_block = PythonTryBlock(node, try_name, self.node_factory,
                                   self.current_function, node.body)
        try_block.sil_name = try_name
        self.stmt_container.labels.append(try_name)
        post_name = self.stmt_container.get_fresh_name('post_with')
        try_block.post_name = post_name
        self.stmt_container.labels.append(post_name)
        finally_name = self.stmt_container.get_fresh_name('with_finally')
        if len(node.items) != 1:
            msg = 'with block may only have one item'
            raise UnsupportedException(node, msg)
        try_block.with_item = node.items[0]
        try_block.finally_name = finally_name
        self.stmt_container.labels.append(finally_name)
        self.stmt_container.try_blocks.append(try_block)

    def visit_Try(self, node: ast.Try) -> None:
        """
        Creates PythonTryBlocks and PythonExceptionHandlers.
        """
        try_name = self.stmt_container.get_fresh_name('try')
        try_block = PythonTryBlock(node, try_name, self.node_factory,
                                   self.stmt_container, node.body)
        try_block.sil_name = try_name
        self.stmt_container.labels.append(try_name)
        post_name = self.stmt_container.get_fresh_name('post_try')
        try_block.post_name = post_name
        self.stmt_container.labels.append(post_name)
        self.stmt_container.try_blocks.append(try_block)
        for handler in node.handlers:
            if handler.type:
                handler_type = self.find_or_create_target_class(handler.type)
            else:
                # Handler has no explicit type, therefore catches any kind
                # of exception.
                handler_type = self.module.global_module.classes['Exception']
            handler_name = self.stmt_container.get_fresh_name(
                'handler' + handler_type.name)
            py_handler = PythonExceptionHandler(handler, handler_type,
                                                try_block, handler_name,
                                                handler.body, handler.name)
            self.stmt_container.labels.append(handler_name)
            try_block.handlers.append(py_handler)
        if node.orelse:
            handler_name = self.stmt_container.get_fresh_name('try_else')
            py_handler = PythonExceptionHandler(node, None, try_block,
                                                handler_name, node.orelse, None)
            try_block.else_block = py_handler
        if node.finalbody:
            finally_name = self.stmt_container.get_fresh_name('try_finally')
            try_block.finally_block = node.finalbody
            try_block.finally_name = finally_name
            self.stmt_container.labels.append(finally_name)
        self.visit_default(node)

    def _incompatible_decorators(self, decorators) -> bool:
        return ((('Predicate' in decorators) and ('Pure' in decorators)) or
                (('IOOperation' in decorators) and (len(decorators) != 1)) or
                (('property' in decorators) and (len(decorators) != 1)) or
                (('AllLow' in decorators) and ('PreservesLow' in decorators)) or
                ((('AllLow' in decorators) or ('PreservesLow' in decorators)) and (
                    ('Predicate' in decorators) or ('Pure' in decorators)))
               )

    def is_declared_contract_only(self, func: ast.FunctionDef) -> bool:
        """
        Checks if the given function is declared to be contract only by the
        respective decorator.
        """
        decorators = {d.id for d in func.decorator_list if isinstance(d, ast.Name)}
        if self._incompatible_decorators(decorators):
            raise InvalidProgramException(func, "decorators.incompatible")
        result = 'ContractOnly' in decorators
        return result

    def is_contract_only(self, func: ast.FunctionDef) -> bool:
        """
        Checks if the given function is supposed to be treated as contract only
        because it is either declared to be so with a decorator, because it's
        not in the main module that's being verified, or because some methods
        were explicitly selected to be verified and this one is not one of them.
        """
        result = self.is_declared_contract_only(func)
        if self.selected:
            selected = False
            if self.current_class:
                if self.current_class.name in self.selected:
                    selected = True
                else:
                    combined = self.current_class.name + '.' + func.name
                    if combined in self.selected:
                        selected = True
            if func.name in self.selected:
                selected = True
            result = result or (not selected)
        return result

    def has_decorator(self, func: ast.FunctionDef, decorator: str) -> bool:
        decorators = {d.id for d in func.decorator_list if isinstance(d, ast.Name)}
        if self._incompatible_decorators(decorators):
            raise InvalidProgramException(func, "decorators.incompatible")
        return decorator in decorators

    def is_pure(self, func: ast.FunctionDef) -> bool:
        return self.has_decorator(func, 'Pure')

    def is_predicate(self, func: ast.FunctionDef) -> bool:
        return self.has_decorator(func, 'Predicate')

    def is_static_method(self, func: ast.FunctionDef) -> bool:
        return self.has_decorator(func, 'staticmethod')

    def is_class_method(self, func: ast.FunctionDef) -> bool:
        return self.has_decorator(func, 'classmethod')

    def is_io_operation(self, func: ast.FunctionDef) -> bool:
        return self.has_decorator(func, 'IOOperation')

    def is_property_getter(self, func: ast.FunctionDef) -> bool:
        return self.has_decorator(func, 'property')

    def is_property_setter(self, func: ast.FunctionDef) -> bool:
        setter_decorator = [d for d in func.decorator_list
                            if isinstance(d, ast.Attribute)]
        if len(setter_decorator) == 0:
            return False
        if len(setter_decorator) > 1 or setter_decorator[0].attr != 'setter':
            raise InvalidProgramException(func, 'unknown.decorator')
        return self.current_class.fields[setter_decorator[0].value.id]

    def is_all_low(self, func: ast.FunctionDef) -> bool:
        return self.has_decorator(func, 'AllLow')

    def preserves_low(self, func: ast.FunctionDef) -> bool:
        return self.has_decorator(func, 'PreservesLow')



"""
This Source Code Form is subject to the terms of the Mozilla Public
License, v. 2.0. If a copy of the MPL was not distributed with this
file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""

"""This module contains all supported obligation types."""


from nagini_translation.translators.obligation.types.must_terminate import (
    MustTerminateObligationInstance,
    MustTerminateObligation,
)
